
import time
import sys
import numpy as np
import pandas as pd
from sklearn.utils import check_random_state
from joblib import Parallel, delayed

from ..random.permute_blocks import block_permutation, get_auto_vg

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.mass_univariate import permuted_ols

def _self_inverse(arr):
    return arr @ np.linalg.pinv(arr)

def _get_contrast(tested_vars, confounding_vars):
    
    dtype = dtype=tested_vars.dtype.name
    
    tv_s = tested_vars.shape[1]
    cv_s = confounding_vars.shape[1]

    contrast = np.vstack([np.eye(tv_s, dtype=dtype),
                          np.zeros((cv_s, tv_s), dtype=dtype)])
    
    return contrast

def _get_perm_matrix(permutation_structure, random_state, intercept_test=False, within_grp=True):

    # If intercept test, ignore permutation structure,
    # and the permutation is just eye with random 1's and -1's.
    if intercept_test:
        flip_vals = (check_random_state(random_state).randint(2, size=(len(permutation_structure))) * 2) - 1
        return np.eye(len(permutation_structure)) * flip_vals
    
    # Gen permutation
    p_set = block_permutation(np.eye(len(permutation_structure), dtype='bool'),
                              blocks=permutation_structure,
                              random_state=random_state,
                              within_grp=within_grp)

    return p_set

def _run_permutation_chunks(run_perm_func, original_scores,
                            thread_id, target_vars, rz, hz,
                            input_matrix, variance_groups, drm,
                            contrast, n_perm_chunk, n_perm, random_state,
                            permutation_structure=None, verbose=0,
                            use_z=False, two_sided_test=True,
                            intercept_test=False,
                            use_special_tf=False,
                            within_grp=True):

    # If the special tensorflow case, and the first job
    if use_special_tf and thread_id == 1:
        from ._tf_run_permutation import run_permutation, cast_input_to_tensor

        # Override the run_perm_func with the tensorflow version
        run_perm_func = run_permutation

        # Cast input to tensor
        target_vars, rz, hz, input_matrix, drm, contrast =\
            cast_input_to_tensor(target_vars, rz, hz, input_matrix,
                                    drm, contrast, dtype=target_vars.dtype)

    # Init starting vars
    t0 = time.time()
    h0_vmax_part = np.empty(n_perm_chunk)
    scores_as_ranks_part = np.zeros(target_vars.shape[1])

    # Run each permutation
    for i in range(n_perm_chunk):
        
        # Generate permutation on the fly
        p_set = _get_perm_matrix(permutation_structure=permutation_structure,
                                 random_state=random_state+i,
                                 intercept_test=intercept_test,
                                 within_grp=within_grp)

        # Get v stats for this permutation
        perm_scores = run_perm_func(
            p_set=p_set, target_vars=target_vars, rz=rz, hz=hz,
            input_matrix=input_matrix, variance_groups=variance_groups,
            drm=drm, contrast=contrast, use_z=use_z)
        
        # Convert perm scores to to absolute values if two sides
        if two_sided_test:
            perm_scores = np.fabs(perm_scores)

        # Add max v stat
        h0_vmax_part[i] = np.nanmax(perm_scores)

        # Find the rank of the original scores in h0_part,
        # basically just compares the vmax and adds 1 if the
        # original score is higher, so e.g., if original score
        # higher than all permutations, will eventually have the highest rank.
        scores_as_ranks_part += (h0_vmax_part[i] < original_scores)
    
        # Per-step verbosity - from nilearn func
        if verbose > 0:
            step = 11 - min(verbose, 10)
            if i % step == 0:
                if n_perm == n_perm_chunk:
                    crlf = "\r"
                else:
                    crlf = "\n"
                percent = float(i) / n_perm_chunk
                percent = round(percent * 100, 2)
                dt = time.time() - t0
                remaining = (100. - percent) / max(0.01, percent) * dt
                sys.stderr.write(
                    "Job #%d, processed %d/%d permutations "
                    "(%0.2f%%, %i seconds remaining)%s"
                    % (thread_id, i, n_perm_chunk, percent, remaining, crlf))

    if verbose > 0:
        print(f'Job finished in {time.time() - t0}')

    return scores_as_ranks_part, h0_vmax_part

def _proc_input(tested_vars, confounding_vars):
    
    # Stack input vars
    input_vars = np.hstack([tested_vars, confounding_vars])
    
    # Get contrasts
    contrast = _get_contrast(tested_vars, confounding_vars)
    contrast_pinv = np.linalg.pinv(contrast.T)
    I_temp = np.eye(len(contrast), dtype=tested_vars.dtype.name)
    contrast0 = I_temp - contrast @ np.linalg.pinv(contrast)
    
    # Selects X - not sure
    # what different is between just
    # using tested_vars directly
    tmpX = input_vars @ contrast_pinv
    
    # Gets svd for confounds
    tmpZ = np.linalg.svd(input_vars @ contrast0)[0]

    # Get Z and X
    Z = tmpZ[:,0:confounding_vars.shape[-1]]
    X = tmpX - Z @ np.linalg.pinv(Z) @ tmpX

    return X, Z, contrast

def _proc_n_perm_chunks(n_perm, n_jobs, use_special_tf=False, special_tf_job_split=None):

    # Special case
    if use_special_tf:

        # Set by default 75% of permutations to run on gpu
        gpu_perms = np.int(np.round(n_perm * special_tf_job_split))

        # Split rest across remaining jobs
        rest_perm_chunks = _proc_n_perm_chunks(n_perm - gpu_perms, n_jobs-1)
        
        # Combine and return
        return np.concatenate([[gpu_perms], rest_perm_chunks]).astype('int')

    # Otherwise, generate n_perm_chunks and return those
    if n_perm > n_jobs:
        n_perm_chunks = np.asarray([n_perm / n_jobs] * n_jobs, dtype=int)

        # Distribute rest to remaining
        remaining = n_perm % n_jobs
        for r in range(remaining):
            n_perm_chunks[r] += 1
    else:
        n_perm_chunks = np.ones(n_perm, dtype=int)

    return n_perm_chunks.astype('int')

def _proc_dtype(vars):

    # Passed as int cases
    if 'int' in vars.dtype.name:

        if vars.dtype.name == 'int64':
            return vars.astype('float64')

        # Default other case
        return vars.astype('float32')

    return vars

def _nan_check(vars):
    if np.isnan(vars).any():
        raise RuntimeError('Currently no missing data is supported in tested_vars, target_vars or confounds_vars.')

def _proc_min_vg_size(min_vg_size, confounding_vars, tested_vars,
                     target_vars, permutation_structure, variance_groups):

    # Calc sizes
    u_groups, u_vg_cnts = np.unique(variance_groups, return_counts=True)

    print(f'Dropping {len(u_groups[u_vg_cnts < min_vg_size])} data points for < min_vg_size = {min_vg_size}', flush=True)

    # Get group index to keep
    to_keep = u_groups[u_vg_cnts >= min_vg_size]

    # Get index of subjects to keep
    i = np.where(np.isin(variance_groups, to_keep))[0]

    return (confounding_vars[i], tested_vars[i], target_vars[i],
            permutation_structure[i], variance_groups[i])

def _process_permuted_v_base(tested_vars,
                             target_vars,
                             confounding_vars,
                             permutation_structure,
                             n_perm=100,
                             variance_groups=None,
                             two_sided_test=True,
                             within_grp=True,
                             random_state=None,
                             use_tf=False,
                             n_jobs=1,
                             dtype=None,
                             use_z=False,
                             model_intercept=False,
                             demean_targets=True,
                             demean_confounds=True,
                             min_vg_size=None):

    # Make sure input okay / right types
    if n_perm < 0:
        raise RuntimeError('n_perm must be 0 or greater.')
    if not n_jobs > 0:
        raise RuntimeError('n_jobs must be 1 or more.')
    if not isinstance(two_sided_test, bool):
        raise RuntimeError('two_sided_test must be passed a bool True / False.')
    if not isinstance(within_grp, bool):
        raise RuntimeError('within_grp must be passed a bool True / False.')

    # Get rng instance from passed random_state
    rng = check_random_state(random_state)

    # Special case of n_jobs > 1 and use_tf
    special_tf_job_split = .75
    if isinstance(use_tf, float):
        special_tf_job_split = use_tf
        use_tf = True

        if special_tf_job_split >= 1 or special_tf_job_split <= 0:
            raise RuntimeError('use_tf must be between 0 and 1 or True / False.')

    # Error if both use_tf and request calculate z
    if use_tf and use_z:
        raise RuntimeError('use_z is not currently supported with tensorflow version.')

    # If confounding vars passed as None
    if confounding_vars is None:
        confounding_vars = np.zeros((len(tested_vars), 0))

    # If passed as DataFrame or Series cast to array
    if isinstance(tested_vars, (pd.DataFrame, pd.Series)):
        tested_vars = np.array(tested_vars)
    if isinstance(target_vars, (pd.DataFrame, pd.Series)):
        target_vars = np.array(target_vars)
    if isinstance(confounding_vars, (pd.DataFrame, pd.Series)):
        confounding_vars = np.array(confounding_vars)
    if isinstance(permutation_structure, (pd.DataFrame, pd.Series)):
        permutation_structure = np.array(permutation_structure)

    # Make sure input shape / dims are correct
    if len(tested_vars.shape) == 1:
        tested_vars = np.expand_dims(tested_vars, axis=1)
   
    if tested_vars.shape[1] > 1:
        raise RuntimeError('You may not pass more than one vars to test!')

    if len(confounding_vars.shape) == 1:
        confounding_vars = np.expand_dims(confounding_vars, axis=1)

    # Make sure the same len for main passed vars
    if (len(tested_vars) != len(target_vars)) or (len(confounding_vars)) !=\
        len(permutation_structure) or (len(confounding_vars) != len(tested_vars)):
        raise RuntimeError('The passed lengths for tested_vars, target_vars, ' \
                           'confounding_vars and permutation structure must all be the same!')
    
    # Check for nans
    _nan_check(confounding_vars)
    _nan_check(tested_vars)
    _nan_check(target_vars)
    _nan_check(permutation_structure)

    # Proc dtype
    confounding_vars = _proc_dtype(confounding_vars)
    tested_vars = _proc_dtype(tested_vars)
    target_vars = _proc_dtype(target_vars)

    # Check if tested_vars is intercept or not
    if np.unique(tested_vars).size == 1:
        intercept_test = True
    else:
        intercept_test = False

    # TODO make sure all the same dtype

    # Optionally add intercept
    if model_intercept and not intercept_test:
        confounding_vars = np.hstack((confounding_vars, np.ones((len(tested_vars), 1))))

    # Custom variance groups passed
    if variance_groups is not None:
        
        # Checks
        if isinstance(variance_groups, (pd.DataFrame, pd.Series)):
            variance_groups = np.array(variance_groups)
        
        _nan_check(variance_groups)

    # Otherwise calculate from permutation structure
    else:
        variance_groups = get_auto_vg(permutation_structure, within_grp=within_grp)

    # Optionally, remove some data points based on too small unique variance group
    if min_vg_size is not None:
        confounding_vars, tested_vars, target_vars, permutation_structure, variance_groups =\
            _proc_min_vg_size(min_vg_size, confounding_vars, tested_vars,
                              target_vars, permutation_structure, variance_groups)

    # De-mean target variable
    if demean_targets:
        target_vars -= target_vars.mean(axis=0)

    # If set False and intercept test, warn
    elif intercept_test:
        print('Warning: Preforming intercept test based on sign flips, if target vars are not de-meaned this can cause problems!', flush=True)

    # Optionally de-mean confounds, can prevent common
    # problems, especially when passing dummy coded variables
    # so default is True
    if demean_confounds:
        confounding_vars -= confounding_vars.mean(axis=0)

    # Process dtype argument if passed
    if dtype is not None:
        tested_vars = tested_vars.astype(dtype)
        confounding_vars = confounding_vars.astype(dtype)
        target_vars = target_vars.astype(dtype)

    # Process / resid ?
    tested_resid, confounding_resid, contrast =\
        _proc_input(tested_vars, confounding_vars)

    # Get all input vars
    input_matrix = np.hstack([tested_resid, confounding_resid])

    # Pre-calculate here
    I = np.eye(len(input_matrix), dtype=input_matrix.dtype.name)
    drm = np.diag(I - _self_inverse(input_matrix))

    # Get inverse of confounding vars
    hz = _self_inverse(confounding_resid)
    rz = I - hz

    # Save some memory
    del tested_resid, tested_vars, confounding_vars, confounding_resid

    # If we are going to use tensorflow
    use_special_tf = False
    if use_tf:

        # If n_jobs is 1, base behavior just run tensorflow
        if n_jobs == 1:

            from ._tf_run_permutation import run_permutation, cast_input_to_tensor

            # Cast to tensor
            target_vars, rz, hz, input_matrix, drm, contrast =\
                cast_input_to_tensor(target_vars, rz, hz, input_matrix,
                                     drm, contrast, dtype=dtype)
        
        # Otherwise, special multi-processing case
        else:

            from ._base_run_permutation import run_permutation
            
            # Special tensorflow / cpu case
            # so we use base cpu permutation for the initial baseline
            # then use the special
            use_special_tf = True
        
    else:

        from ._base_run_permutation import run_permutation

    # Get original scores, pass permutation as None
    original_scores = run_permutation(p_set=np.eye(len(input_matrix)),
                                      target_vars=target_vars,
                                      rz=rz, hz=hz, input_matrix=input_matrix,
                                      variance_groups=variance_groups,
                                      drm=drm, contrast=contrast,
                                      use_z=use_z)

    # If two sided test, get absolute value of the scores
    original_scores_sign = None
    if two_sided_test:
        original_scores_sign = np.sign(original_scores)
        original_scores = np.fabs(original_scores)

    # A little ugly but ... for now
    return (original_scores, original_scores_sign,
            run_permutation, target_vars,
            rz, hz, input_matrix, permutation_structure, variance_groups, 
            drm, contrast, rng, intercept_test, use_special_tf, special_tf_job_split)

def permuted_v(tested_vars,
               target_vars,
               confounding_vars=None,
               permutation_structure=None,
               within_grp=True,
               n_perm=100,
               variance_groups=None,
               two_sided_test=True,
               demean_targets=True,
               demean_confounds=True,
               model_intercept=False,
               min_vg_size=None,
               dtype=None,
               use_tf=False,
               use_z=False,
               random_state=None,
               n_jobs=1,
               verbose=0):
    '''This function is used to perform a permutation based statistical test
    on data with an underlying exchangability-block type structure.

    In the case that no permutation structure is passed, i.e., value of None,
    then the original function will NOT be called. Instead,
    :func:`nilearn.mass_univariate.permuted_ols` will be used instead, and
    t-statistics calculated!

    In the case that the passed tested_vars are a single group only, e.g., all
    1's, then any passed permutation_structure will only be used to generate variance groups,
    as instead of permutations based on swapping data, permutations will be performed
    through random sign flips of the data.

    This code is based to a large degree upon matlab code from
    program PALM, as well as influence by the permutation function in python
    library nilearn.

    Parameters
    -----------
    tested_vars : numpy array or pandas DataFrame
        Pass as an array or DataFrame, with shape
        either 1D or 2D as subjects x 1, and containing the
        values of interest in which to calculate the v-statistic
        against the passed `target_vars`, and as corrected for
        any passed `confounding_vars`.

        Note: The subject / data point order and length should match exactly
        between the first dimensions of `tested_vars`, 'target_vars`,
        `confounding_vars` and `permutation_structure`.

    target_vars : numpy array
        A 2D numpy array with shape subjects x features, containing
        the typically imaging features per subject to univariately
        calculate v-stats for.

        Note: The subject / data point order and length should match exactly
        between the first dimensions of `tested_vars`, 'target_vars`,
        `confounding_vars` and `permutation_structure`.

    confounding_vars : numpy array, pandas DataFrame or None, optional

        Confounding variates / covariates, passed as a 2D numpy array
        with shape subjects x number of confounding variables, or as None.
        If None, no variates are added, except maybe a constant column according to the
        value of parameter `model_intercept`. Otherwise, the passed variables influence
        will be removed from `tested_vars` before calculating the relationship between
        `tested_vars` and `target_vars`.

        Note: The subject / data point order and length should match exactly
        between the first dimensions of `tested_vars`, 'target_vars`,
        `confounding_vars` and `permutation_structure`.

        ::

            default = None

    permutation_structure : numpy array, pandas DataFrame or None, optional
        This parameter represents the underlying exchangability-block
        structure of the data passed. It is also used in order to automatically determine
        the underlying variance structure of the passed data.

        See PALM's documentation for an introduction on how to format ExchangeabilityBlocks:
        https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/PALM/ExchangeabilityBlocks

        This parameter accepts the same style input as PALM, except it is passed
        here as an array or DataFrame instead of as a file. The main requirement 
        is that the shape of the structure match the number of subjects / data points
        in the first dimension.

        In this case that no permutation structure is passed, i.e., value of None,
        then the original function will NOT be called. Instead,
        :func:`nilearn.mass_univariate.permuted_ols` will be used instead, and
        t-statistics calculated!
        
        Note: The subject / data point order and length should match exactly
        between the first dimensions of `tested_vars`, 'target_vars`,
        `confounding_vars` and `permutation_structure`.

        ::

            default = None

    within_grp : bool, optional
        This parameter is only relevant when a permutation structure is passed, in that
        case it describes how the left-most exchanability / permutation structure column should act.
        Specifically, if True, then it specifies that the left-most column should be treated as groups
        to act in a within group swap only manner. If False, then it will consider the left-most column
        groups to only be able to swap at the group level with other groups of the same size.

        ::

            default = True

    n_perm : int, optional
        The number of permutations to perform. Permutations are costly but the more are performed,
        the more precision one gets in the p-values estimation.

        If passed 0, this a valid option, and can be used to calculate just the original scores.

        ::

            default = 500

    variance_groups : None or numpy array / pandas DataFrame, optional
        By default these are automatically computed from the passed permutation structure.
        Otherwise, this parameter can be overriden by passing a custom set of variance groups.

        ::

            default = None

    two_sided_test : bool, optional

        If True, performs an unsigned v-test, where
        both positive and negative effects are considered.
        If False, only positive effects are considered as relevant.
        
        ::

            default = True

    demean_targets : bool, optional
        If True, then the passed `target_vars` are demeaned across passed
        subjects (i.e., each single feature scaled to have mean 0 across subjects).

        Note: If performing an intercept based test (i.e., the tested vars are all 1's)
        then this parameter should be left True, as sign flips to data that are not de-meaned
        might cause strange issues.

        ::

            default = True

    demean_confounds : bool, optional
        If True, then the passed `confounding_vars` are demeaned across passed
        subjects (i.e., each single variable / column scaled to have mean 0 across subjects).

        ::
        
            default = True

    model_intercept : bool, optional
      If True, a constant column is added to the confounding variates
      unless the tested variate is already the intercept.
      
      ::

        default = False

    min_vg_size : int or None, optional
        If None, this parameter is ignored. Otherwise,
        if passed as int, this defines the smallest sized unique variance
        group allowed. Specifically, variance groups are calculated automatically
        from the passed permutation structure, and like-wise some statistics are calculated
        seperately per variance group, so that it can be a group idea to set a filter here
        that will drop any subject's data that are below that threshold.

        ::

            default = None
        
    dtype : str or None, optional
        If left as default of None, then the original
        datatypes for all passed data will be used.
        Alternatively, you may specify either 'float32' or 'float64'
        and all data and calculations will be cast to that datatype.

        It can be very beneficial in practice to specify 'float32'
        and perform less precise calculations. This can greatly improve
        memory and will also provide a speedup (a significant speed up when
        `use_tf` is set, otherwise a modest one).

        ::

            default = None
    
    use_tf : bool, optional
        This flag specifies if permutations should be
        run on a special optimized version of the code
        designed to use a GPU, written in tensorflow.

        Note: If True, then this parameter requires that you have
        tensorflow installed and working. Ideally, also setup with a gpu,
        as using tensorflow with a cpu will not provide any benefit relative to using
        the base numpy version of the code.

        ::

            default = False

    use_z : bool, optional
        v-statstics can optionally be converted into z-statstics. If
        passed True, then the returned `original_scores` will be z-statistics
        instead of v-statstics, and likewise, the permutation test will be performed
        by comparing max z-stats instead of v-stats.

        Note: This option cannot be used with use_tf.

        ::

            default = False
    
    random_state : int or None, optional
        The random seed for random number generator. This
        can be set by passing an int to specify the same permutations, or
        by passing None a random seed will be used.

        ::

            default = None
    
    n_jobs : int, optional

        Number of parallel workers. If 0 is provided, all CPUs are used.
        A negative number indicates that all the CPUs except (abs(n_jobs) - 1) ones will be used.

        ::

            default = 1
    
    verbose : int, optional
        verbosity level (0 means no message).

        ::

            default = 0

    Returns
    --------
    pvals : numpy array
        The p-values associated with the significance test of the
        explanatory variates against the target variates. Family-wise corrected p-values.
    
    original_scores : numpy array
        Either v (default), z or t statistics associated with the
        significance test of the explanatory variates against the target variates.
        The ranks of the scores into the h0 distribution correspond to the p-values.

    h0_vmax : numpy array
        Distribution of the (max) v/z/t-statistic under the null hypothesis
        (obtained from the permutations). Array is sorted.

    '''

    # Special case, switch functions
    if permutation_structure is None:
        print('No permutation structure passed, switching to nilearn permuted_ols!', flush=True)
        return permuted_ols(tested_vars=tested_vars, target_vars=target_vars,
                            confounding_vars=confounding_vars, model_intercept=model_intercept,
                            n_perm=n_perm, two_sided_test=two_sided_test, random_state=random_state,
                            n_jobs=n_jobs, verbose=verbose)

    # TODO add handle missing-ness better?
    # TODO if tested vars multiple, setup to run each seperately ... same as nilearn
    # TODO change to class based for cleaner? Whole function as a class instead
    # TODO support no confounding vars (in that case add intercept?)

    # Separate function for initial processing
    (original_scores, original_scores_sign, run_permutation,
     target_vars, rz, hz, input_matrix, permutation_structure,
     variance_groups, drm, contrast, rng, intercept_test,
     use_special_tf, special_tf_job_split) =\
        _process_permuted_v_base(tested_vars=tested_vars,
                                 target_vars=target_vars,
                                 confounding_vars=confounding_vars,
                                 permutation_structure=permutation_structure,
                                 n_perm=n_perm,
                                 variance_groups=variance_groups,
                                 two_sided_test=two_sided_test,
                                 within_grp=within_grp,
                                 random_state=random_state,
                                 use_tf=use_tf,
                                 n_jobs=n_jobs,
                                 dtype=dtype,
                                 use_z=use_z,
                                 model_intercept=model_intercept,
                                 demean_targets=demean_targets,
                                 demean_confounds=demean_confounds,
                                 min_vg_size=min_vg_size)

    # Return original scores - if no permutation
    if n_perm == 0:

        # Convert back to signed scores before returning
        if two_sided_test:
            original_scores *= original_scores_sign

        return np.asarray([]), original_scores, np.asarray([])

    # Get n_perm_chunks, or also case where is n_perm
    # is pre-generated, splits it by jon
    n_perm_chunks = _proc_n_perm_chunks(n_perm, n_jobs,
                                        use_special_tf,
                                        special_tf_job_split)

    # Submit permutations with joblib
    ret = Parallel(n_jobs=n_jobs, verbose=verbose)(delayed(_run_permutation_chunks)(
        run_perm_func=run_permutation, original_scores=original_scores,
        thread_id=thread_id+1, target_vars=target_vars,
        rz=rz, hz=hz, input_matrix=input_matrix,
        variance_groups=variance_groups, drm=drm, contrast=contrast,
        n_perm_chunk=n_perm_chunk, n_perm=n_perm,
        random_state=rng.randint(1, np.iinfo(np.int32).max - 1),
        permutation_structure=permutation_structure,
        verbose=verbose, use_z=use_z, two_sided_test=two_sided_test,
        intercept_test=intercept_test,
        use_special_tf=use_special_tf,
        within_grp=within_grp)
        for thread_id, n_perm_chunk in enumerate(n_perm_chunks))

    # Collect returned results together
    scores_as_ranks_parts, h0_vmax_parts = zip(*ret)

    # Stack together the max v stats from each permutation
    h0_vmax = np.hstack(h0_vmax_parts)

    # Combine the scores as ranks
    scores_as_ranks = np.sum(scores_as_ranks_parts, axis=0)
 
    # Convert ranks into p-values
    pvals = (n_perm + 1 - scores_as_ranks) / float(1 + n_perm)

    # Convert back to signed scores before returning
    if two_sided_test:
        original_scores *= original_scores_sign

    return pvals, original_scores, h0_vmax



"""
PCA dimension reduction on multiple subjects.
This is a good initialization method for ICA.
"""
import numpy as np
from joblib import Memory
from sklearn.utils.extmath import randomized_svd

from .base import BaseDecomposition

class MultiPCA(BaseDecomposition):

    def __init__(self, n_components=20,
                 do_cca=True,
                 random_state=None,
                 n_jobs=1,
                 verbose=0):
    
        self.n_components = n_components
        self.do_cca = do_cca

        BaseDecomposition.__init__(self, n_components=n_components,
                                   random_state=random_state,
                                   n_jobs=n_jobs,
                                   verbose=verbose)

    def _raw_fit(self, data):
        """Helper function that directly process unmasked data"""

        if self.do_cca:
            S = np.sqrt(np.sum(data ** 2, axis=1))
            S[S == 0] = 1
            data /= S[:, np.newaxis]

        components_, self.variance_, _ = randomized_svd(
            data.T, n_components=self.n_components,
            transpose=True,
            random_state=self.random_state, n_iter=3)
        
        if self.do_cca:
            data *= S[:, np.newaxis]
        self.components_ = components_.T
    
        return components_


"""
CanICA
"""

# Author: Alexandre Abraham, Gael Varoquaux,
# License: BSD 3 clause

import warnings as _warnings
import numpy as np

from operator import itemgetter

from scipy.stats import scoreatpercentile
from sklearn.decomposition import fastica
from joblib import Memory, delayed, Parallel
from sklearn.utils import check_random_state

from .multi_pca import MultiPCA

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn._utils import fill_doc


@fill_doc
class CanICA(MultiPCA):

    def __init__(self,
                 n_components=20,
                 do_cca=True,
                 threshold='auto',
                 n_init=10,
                 random_state=None,
                 n_jobs=1, verbose=0
                 ):

        super(CanICA, self).__init__(
            n_components=n_components,
            do_cca=do_cca,
            random_state=random_state,
            n_jobs=n_jobs, verbose=verbose)

        if isinstance(threshold, float) and threshold > n_components:
            raise ValueError("Threshold must not be higher than number "
                             "of maps. "
                             "Number of maps is %s and you provided "
                             "threshold=%s" %
                             (str(n_components), str(threshold)))
        self.threshold = threshold
        self.n_init = n_init

    def _unmix_components(self, components):
        """Core function of CanICA than rotate components_ to maximize
        independence"""
        random_state = check_random_state(self.random_state)

        seeds = random_state.randint(np.iinfo(np.int32).max, size=self.n_init)
        
        # Note: fastICA is very unstable, hence we use 64bit on it
        results = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)(
            delayed(fastica)
            (components.astype(np.float64), whiten=True, fun='cube',
             random_state=seed)
            for seed in seeds)

        ica_maps_gen_ = (result[2].T for result in results)
        ica_maps_and_sparsities = ((ica_map,
                                    np.sum(np.abs(ica_map), axis=1).max())
                                   for ica_map in ica_maps_gen_)
        ica_maps, _ = min(ica_maps_and_sparsities, key=itemgetter(-1))

        # Thresholding
        ratio = None
        if isinstance(self.threshold, float):
            ratio = self.threshold
        elif self.threshold == 'auto':
            ratio = 1.
        elif self.threshold is not None:
            raise ValueError("Threshold must be None, "
                             "'auto' or float. You provided %s." %
                             str(self.threshold))
        if ratio is not None:
            abs_ica_maps = np.abs(ica_maps)
            percentile = 100. - (100. / len(ica_maps)) * ratio
            if percentile <= 0:
                _warnings.warn("Nilearn's decomposition module "
                               "obtained a critical threshold "
                               "(= %s percentile).\n"
                               "No threshold will be applied. "
                               "Threshold should be decreased or "
                               "number of components should be adjusted." %
                               str(percentile), UserWarning, stacklevel=4)
            else:
                threshold = scoreatpercentile(abs_ica_maps, percentile)
                ica_maps[abs_ica_maps < threshold] = 0.
        # We make sure that we keep the dtype of components
        self.components_ = ica_maps.astype(self.components_.dtype)

        # flip signs in each component so that peak is +ve
        for component in self.components_:
            if component.max() < -component.min():
                component *= -1

    # Overriding MultiPCA._raw_fit overrides MultiPCA.fit behavior
    def _raw_fit(self, data):
        """Helper function that directly process unmasked data.
        Useful when called by another estimator that has already
        unmasked data.
        Parameters
        ----------
        data : ndarray or memmap
            Unmasked data to process
        """
        components = MultiPCA._raw_fit(self, data)
        self._unmix_components(components)
        return self


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from .plot_single_surf import plot_single_surf, add_collage_colorbar
from .ref import SurfRef
from ..transform.space import process_space
from scipy.stats import scoreatpercentile
import nibabel as nib
from ..misc.print import _get_print
import matplotlib.colors as colors

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.plotting import plot_glass_brain, plot_stat_map, plot_roi


def _trunc_cmap(cmap, minval=0.0, maxval=1.0, n=1000):
    
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    
    return new_cmap

def _proc_vs(data, vmin, vmax, symmetric_cbar):
    
    # If already set, skip
    if vmin is not None and vmax is not None:
        return vmin, vmax
    
    # Get data as flat array
    flat_data = _collapse_data(data)

    # Small percent of min / max to add to either end
    s = np.nanmax(np.abs(flat_data)) / 25
    
    # Both not set case
    if vmin is None and vmax is None:

        # If not symmetric_cbar, then these value
        # stay fixed as this
        vmin, vmax = np.nanmin(flat_data) - s, np.nanmax(flat_data) + s
        
        # If symmetric need to override either
        # vmin or vmax to be the opposite of the larger
        if symmetric_cbar:
            
            # vmin is larger, override vmax
            if np.abs(vmin) > vmax:
                vmax = -vmin
            
            # vmax is larger, override vmin
            else:
                vmin = -vmax
    
    # If just vmin not set
    if vmin is None:
        
        # If vmax set, vmin not, and symmetric cbar
        # then vmin is -vmax
        if symmetric_cbar:
            vmin = -vmax
        
        # Otherwise, vmin is just the min value in the data
        else:
            vmin = np.nanmin(flat_data) - s

    # If vmax not set, same cases as above but flipped
    if vmax is None:

        if symmetric_cbar:
            vmax = -vmin
        else:
            vmax = np.nanmax(flat_data) + s

    return vmin, vmax


def _proc_threshold(data, threshold, percentile=75, rois=False):
    
    # If None, set to a bit below minimum
    if threshold is None:
        flat_abs_data = np.abs(_collapse_data(data))

        # Want to handle 0's and NaN's
        return np.nanmin(flat_abs_data[np.where(flat_abs_data != 0)]) / 5

    # Only proc if left at auto
    if not (threshold == 'auto' or threshold == 'default'):
        return threshold

    # If ROIs then default is .5
    # to not plot rois marked 0
    if rois:
        return .5
    
    # Get flat data
    flat_data = _collapse_data(data)
    
    # Get percentile - not counting any zeros!
    threshold = scoreatpercentile(flat_data[flat_data != 0], per=percentile) - 1e-5

    return threshold

def _collapse_data(data):
    '''Assumes data is in standard {} form.'''
    
    # If directly nifti image
    if isinstance(data, nib.Nifti1Image):
        return np.array(data.get_fdata()).flatten()
    
    # Directly ndarray case
    elif isinstance(data, np.ndarray):
        return np.array(data).flatten()
    
    # Init empty list
    collapsed = []
    
    # Handle passed as dict or list
    # of arrays
    for key in data:

        # Get as item / array
        if isinstance(data, dict):
            item = data[key]
        else:
            item = key

        # Recursive case add
        collapsed.append(_collapse_data(item))

    # Return as concat version
    return np.concatenate(collapsed)


def _get_if_sym_cbar(data, symmetric_cbar, rois=False):
    '''Assumes data is in standard {} form.'''
     
    # If user passed, keep that value
    if not (symmetric_cbar == 'auto' or symmetric_cbar == 'default'):
        return symmetric_cbar
    
    # If rois, default = False, so return
    if rois:
        return False
    
    # Get data as 1D array
    flat_data = _collapse_data(data)

    # If all positive or negative, assume false
    if np.all(flat_data >= 0) or np.all(flat_data <= 0):
        return False

    # Otherwise, assume is symmetric
    return True

def _add_plots_to_axes(figure, grid, proj_3d, n_cols, n_rows, widths):

    # If passed as single str,  or None, set for all
    if isinstance(proj_3d, str) or proj_3d is None:
        proj_3d = [[proj_3d for _ in range(n_cols)] for _ in range(n_rows)]

    # Otherwise, we use the grid to add_subplots generating axes
    # with each subplot according to the passed proj_3d if it is
    # 3D or not. Use the original number of passed n_rows and n_cols
    axes = []
    for i in range(n_rows):
        for j in range(n_cols):

            if n_rows == 1:
                axes.append(figure.add_subplot(grid[j], projection=proj_3d[i][j]))
            elif len(widths) == 1:
                axes.append(figure.add_subplot(grid[i], projection=proj_3d[i][j]))
            else:
                axes.append(figure.add_subplot(grid[i, j], projection=proj_3d[i][j]))

    return figure, axes


def _setup_fig_axes(widths, heights, 
                    figure=None, subplot_spec=None,
                    get_grid=False, figsize=(10, 10),
                    proj_3d=None, title=None, title_sz=12,
                    title_y=None, colorbar=False, colorbar_params=None,
                    wspace=1, hspace=1):

    # Keep track of the number of original
    # passed cols and rows.
    if widths is None:
        n_cols = 1
    else:
        n_cols = len(widths)
        widths = widths.copy()

    if heights is None:
        n_rows = 1
    else:
        n_rows = len(heights)

    # Proc arg
    if colorbar_params is None:
        colorbar_params = {}

    # Init colorbar ax as None
    colorbar_ax = None

    # If no figure, init
    if figure is None:
        figure = plt.figure(figsize=figsize)

    # If colorbar, add extra width
    if colorbar is True:
        
        if 'cbar_fig_ratio' in colorbar_params:
            widths += [colorbar_params['cbar_fig_ratio']]
        else:
            widths += [.5]

    # Two cases for init'ing grid, either init as nested
    # if subplot_spec, or as base / new
    if subplot_spec is None:
        grid = gridspec.GridSpec(n_rows, len(widths),
                                 wspace=wspace,
                                 hspace=hspace,
                                 width_ratios=widths,
                                 height_ratios=heights)
    else:
        grid =\
            gridspec.GridSpecFromSubplotSpec(n_rows, len(widths),
                                             wspace=wspace,
                                             hspace=hspace,
                                             width_ratios=widths,
                                             height_ratios=heights,
                                             subplot_spec=subplot_spec)
    
    # Next, add the colorbar axis if colorbar, adds by default
    # slash as only option now to the far right
    if colorbar:

        if n_rows == 1:
            colorbar_ax = figure.add_subplot(grid[-1])
        else:
            colorbar_ax = figure.add_subplot(grid[:, -1])
        
        # Make sure to hide axis here
        colorbar_ax.set_axis_off()

    # If a title is passed, also add a special title axis
    if title is not None:
        
        # Colorbar and title case
        if colorbar:
            if n_rows == 1:
                title_ax = figure.add_subplot(grid[:-1])
            else:
                title_ax = figure.add_subplot(grid[:,:-1])

        # Just title
        else:
            title_ax = figure.add_subplot(grid[:])
            
        # Set title and axis off - (hiding)
        title_ax.set_title(title, fontsize=title_sz, y=title_y)
        title_ax.set_axis_off()

    # Optionally, return here, the grid object and figure directly
    if get_grid:
        return figure, grid, colorbar_ax

    # Otherwise add plots to axes
    figure, axes = _add_plots_to_axes(figure, grid, proj_3d,
                                      n_cols, n_rows, widths)

    # Return each, axes instead of grid
    return figure, axes, colorbar_ax


def plot_surf_hemi(data, ref, hemi,
                   surf_mesh='inflated',
                   bg_map='sulc',
                   symmetric_cbar='auto',
                   vmin=None, vmax=None,
                    **kwargs):

    # Proc if sym cbar auto - assume rois False here
    symmetric_cbar = _get_if_sym_cbar(data, symmetric_cbar, rois=False)
    
    # Proc if sym colorbar
    vmin, vmax = _proc_vs(data, vmin, vmax, symmetric_cbar)

    if hemi == 'lh':
        hemi = 'left'
    if hemi == 'rh':
        hemi = 'right'

    # These get surf functions are pretty smart
    # and tolerant to different inputs
    surf_mesh = ref.get_surf(surf_mesh, hemi)
    bg_map = ref.get_surf(bg_map, hemi)

    return plot_single_surf(surf_mesh=surf_mesh,
                            surf_map=data,
                            bg_map=bg_map,
                            hemi=hemi,
                            vmin=vmin,
                            vmax=vmax,
                            **kwargs)


def plot_surf_collage(data, ref=None, surf_mesh='inflated',
                      bg_map='sulc', view='standard',
                      title=None, title_sz=18, title_y=None,
                      vmin=None, vmax=None,
                      cbar_vmin=None, cbar_vmax=None,
                      figsize=(15, 10), symmetric_cbar='auto',
                      figure=None, subplot_spec=None,
                      wspace=-.35, hspace=-.1,
                      colorbar=False, dist=6.5, b_sz=.5,
                      colorbar_params=None, **kwargs):
    '''
    data should be list with two elements [lh, rh] data.

    view: 'standard' means plot left and right hemisphere lateral
    and medial views in a grid. 
    'fb' / front back, means plot anterior posterior views in a row.
    '''

    # Proc defaults for colorbar params
    colorbar_params = _sort_colorbar_kwargs(colorbar_params=colorbar_params, **kwargs)

    # Proc if sym cbar auto - assume rois False here
    symmetric_cbar = _get_if_sym_cbar(data, symmetric_cbar, rois=False)
    
    # Get vmin and vmax based off data + passed args
    vmin, vmax = _proc_vs(data, vmin, vmax, symmetric_cbar)
    smfs = []
    
    # Set params based on requested view
    if view in ['standard', 'default']:
        hemis = ['lh', 'rh', 'lh', 'rh']
        views = ['lateral', 'lateral', 'medial', 'medial']
        widths, heights = [1, 1], [1, 1]

    elif view in ['flat standard', 'fs', 'row standard']:
        hemis = ['lh', 'rh', 'b', 'lh', 'rh']
        views = ['lateral', 'lateral', 'b', 'medial', 'medial']
        widths, heights = [1, 1, b_sz, 1, 1], [1]

    # TODO still looks like ass
    elif view in ['flat extra', 'fe', 'row extra']:
        hemis = ['lh', 'rh', 'lh', 'rh', 'b', 'lh', 'rh', 'lh', 'rh']
        views = ['anterior', 'anterior', 'lateral', 'lateral', 'b',
                 'medial', 'medial', 'posterior', 'posterior']
        widths, heights = [1, 1, 1, 1, b_sz, 1, 1, 1, 1], [1]

    elif view in ['fb', 'front back', 'ap', 'anterior posterior']:
        hemis = ['lh', 'rh', 'b', 'lh', 'rh']
        views = ['anterior', 'anterior', 'b',
                 'posterior', 'posterior']
        widths, heights = [1, 1, b_sz, 1, 1], [1]

    elif view in ['f', 'front', 'a', 'anterior']:
        hemis = ['lh', 'rh']
        views = ['anterior', 'anterior']
        widths, heights = [1, 1], [1]

    elif view in ['b', 'back', 'p', 'posterior']:
        hemis = ['lh', 'rh']
        views = ['posterior', 'posterior']
        widths, heights = [1, 1], [1]

    # Setup figres and axes
    figure, axes, colorbar_ax =\
        _setup_fig_axes(widths=widths, heights=heights,
                        figure=figure, subplot_spec=subplot_spec,
                        get_grid=False, figsize=figsize,
                        proj_3d='3d', title=title, title_sz=title_sz, title_y=title_y,
                        colorbar=colorbar, colorbar_params=colorbar_params,
                        wspace=wspace, hspace=hspace)

    # Fill the axes with appropriate values
    for i in range(len(hemis)):

        if hemis[i] == 'lh':
            d = 0
        elif hemis[i] == 'rh':
            d = 1
        else:
            axes[i].patch.set_alpha(0)
            axes[i].set_axis_off()
            continue

        figure, axes[i], smf =\
            plot_surf_hemi(data=data[d], ref=ref, hemi=hemis[i],
                           surf_mesh=surf_mesh, bg_map=bg_map,
                           vmin=vmin, vmax=vmax, view=views[i],
                           colorbar=False, dist=dist,
                           figure=figure, axes=axes[i], **kwargs)
        smfs.append(smf)

    # Set to concat
    smfs = np.concatenate(smfs)
    
    # Add color bar
    if colorbar is True:
        figure, colorbar_ax =\
            add_collage_colorbar(
             figure=figure, ax=colorbar_ax, smfs=smfs,
             vmin=vmin, vmax=vmax,
             cbar_vmin=cbar_vmin, cbar_vmax=cbar_vmax,
             multicollage=True,
             colorbar_params=colorbar_params, **kwargs)

    return figure, axes, smfs


def plot_surf_vol_collage(surf, vol,
                          vol_plot_type='glass',
                          cmap='default',
                          threshold='auto',
                          surf_to_vol_ratio=1,
                          vmin=None, vmax=None,
                          cbar_vmin=None, cbar_vmax=None,
                          figure=None,
                          subplot_spec=None,
                          figsize=(20, 20),
                          title=None, title_sz=18,
                          title_y=None,
                          colorbar=False,
                          symmetric_cbar=False,
                          hspace=0, wspace=0,
                          colorbar_params=None,
                          surf_params=None,
                          vol_params=None,
                          verbose=0, _print=None):

    # Proc defaults
    if colorbar_params is None:
        colorbar_params = {}
    if surf_params is None:
        surf_params = {}
    if vol_params is None:
        vol_params = {}

    # Get verbose print object
    _get_print(verbose, _print=_print)

    # Grab data as list
    data_as_list = [surf[0], surf[1], vol.get_fdata()]

    # Process vmin / vmax from data and passed args
    vmin, vmax = _proc_vs(data_as_list, vmin=vmin, vmax=vmax,
                          symmetric_cbar=symmetric_cbar)

    # Proc threshold if auto
    rois = vol_plot_type == 'roi'
    threshold = _proc_threshold(data_as_list, threshold, rois=rois)

    # Init smfs with min and max of vol data, as
    # these values won't change in
    # in the same way plotting surf vals can.
    smfs = [np.array([np.nanmin(vol.get_fdata()), np.nanmax(vol.get_fdata())])]
    
    # Init settings
    widths = [1]
    heights = [surf_to_vol_ratio] + [1]

    # Setup figures get returned as grid instead of axes
    figure, grid, colorbar_ax =\
        _setup_fig_axes(widths=widths,  heights=heights,
                        figure=figure, subplot_spec=subplot_spec,
                        get_grid=True, figsize=figsize, title=title,
                        title_sz=title_sz, title_y=title_y, colorbar=colorbar,
                        colorbar_params=colorbar_params,
                        wspace=wspace, hspace=hspace)

    # Either multi-index w/ colorbar or just two rows w/o
    if colorbar:
        surf_grid, vol_grid = grid[0, 0], grid[1, 0]
    else:
        surf_grid, vol_grid = grid[0], grid[1]

    # Plot surf collage in the top grid spot
    # The already concatenated smf is returned
    figure, _, smf = plot_surf_collage(surf, 
                                       figure=figure,
                                       subplot_spec=surf_grid,
                                       cmap=cmap,
                                       threshold=threshold,
                                       vmin=vmin,
                                       vmax=vmax,
                                       colorbar=False, # Fixed False
                                       **surf_params)
    
    # Keep track of vals for plotting colorbar
    smfs.append(smf)

    # Plot volume
    vol_ax = figure.add_subplot(vol_grid)

    # Process vol_plot_type
    if vol_plot_type == 'glass':
        vol_plot_func = plot_glass_brain
    elif vol_plot_type == 'stat':
        vol_plot_func = plot_stat_map
    elif vol_plot_type == 'roi':
        vol_plot_func = plot_roi
    else:
        raise RuntimeError('vol_plot_type must be one of "glass", "stat" or "roi".')
    
    # Handle funny case where plotting glass
    # but the plotted data is symmetric_cbar and plot abs not specified
    if vol_plot_type == 'glass' and 'plot_abs' not in vol_params:

        # If symmetric cbar, then we want
        # glass to not plot abs
        if symmetric_cbar:
            vol_params['plot_abs'] = False
            _print('Plotting glass brain with plot_abs=False.', level=1)
 
    # Call base plot function
    vol_plot_func(vol,
                  figure=figure, axes=vol_ax, 
                  cmap=cmap, threshold=threshold,
                  vmax=vmax, colorbar=False,
                  **vol_params)

    # Set to concat
    smfs = np.concatenate(smfs)

    # Add color bar
    if colorbar is True:
        figure, colorbar_ax =\
            add_collage_colorbar(
             figure=figure, ax=colorbar_ax, smfs=smfs,
             vmin=vmin, vmax=vmax,
             cbar_vmin=cbar_vmin, cbar_vmax=cbar_vmax,
             multicollage=True,
             colorbar_params=colorbar_params, cmap=cmap,
             threshold=threshold)

    return figure, grid, smfs

## Smart Plot Functions ##

def _proc_ref_arg_defaults(ref, surf_mesh, bg_map, darkness):

    # Allow any user passed args to override space ref defaults here
    if surf_mesh is None or surf_mesh == 'auto' or surf_mesh == 'default':
        surf_mesh = ref.surf_mesh
    if bg_map is None or bg_map == 'auto' or bg_map == 'default':
        bg_map = ref.bg_map
    if darkness is None or darkness == 'auto' or darkness == 'default':
        darkness = ref.darkness

    return surf_mesh, bg_map, darkness


def _load_data_and_ref(data, space=None, hemi=None, _print=None):
    
    # Process the data + space - returns data as dict
    data, space = process_space(data, space, hemi, _print=_print)
    
    # If no space, i.e., just sub, return no Ref
    if space is None:
        return data, None

    # Otherwise generate SurfRef with defaults
    ref = SurfRef(space=space, parc=None)
    
    # If assumed native space, then assume user will pass mesh + bg_map
    if space == 'native':
        ref.surf_mesh = None
        ref.bg_map = None
        ref.darkness = 1
    
    # fs_LR space defaults
    elif 'fs_LR' in space:
        ref.surf_mesh = 'very_inflated'
        ref.bg_map = 'sulc'
        ref.darkness = .5
    
    # Freesurfer spaces defaults
    else:
        ref.surf_mesh = 'inflated'
        ref.bg_map = 'sulc'
        ref.darkness = 1

    return data, ref


def _proc_avg_method(data, rois, avg_method):

    if avg_method is None:
        return None

    # TODO need volumetric case here?

    # If rois force avg method to median
    if rois:
        return 'median'
    
    # Otherwise set based on if we think we are
    # plotting roi values on a surface
    if avg_method == 'default' or avg_method == 'auto':
        
        flat_data = _collapse_data(data)

        # Get flat non-zero
        flat_non_zero = flat_data[flat_data != 0]
        n_unique = len(np.unique(flat_non_zero))

        # If more than 5% data points different
        if n_unique / len(flat_non_zero) > 0.05:
            avg_method = 'mean'
        else:
            avg_method = 'median'

    return avg_method

def _proc_colorbar(colorbar, rois):

    if colorbar == 'default' or colorbar == 'auto':
        colorbar = True
        
        if rois:
            colorbar = False
    
    return colorbar


def _prep_base_auto_defaults(data_as_list, rois,
                             symmetric_cbar, threshold,
                             cmap, colorbar, avg_method):
            
    # Collapse data
    flat_data = _collapse_data(data_as_list)
    
    # Process automatic symmetric_cbar
    symmetric_cbar = _get_if_sym_cbar(flat_data,
                                      symmetric_cbar,
                                      rois=rois) 

    # Proc threshold if auto
    threshold = _proc_threshold(flat_data, threshold, rois=rois)
    
    # Get cmap based on passed + if rois or not + if sym
    cmap = _proc_cmap(cmap, rois, symmetric_cbar, flat_data=flat_data)
    
    # Proc colorbar option - yes if not rois
    colorbar = _proc_colorbar(colorbar, rois)

    # Proc avg method default
    avg_method = _proc_avg_method(flat_data, rois, avg_method)

    # Return
    return symmetric_cbar, threshold, cmap, colorbar, avg_method


def _prep_auto_defaults(data, space, hemi, rois,
                        symmetric_cbar, threshold,
                        cmap, colorbar, avg_method, _print):

    # Fine passing already proc'ed whatever here
    data, ref = _load_data_and_ref(data, space=space, hemi=hemi, _print=_print)

    # Process
    symmetric_cbar, threshold, cmap, colorbar, avg_method =\
        _prep_base_auto_defaults(data, rois,
                                 symmetric_cbar, threshold,
                                 cmap, colorbar, avg_method)

    return data, ref, symmetric_cbar, threshold, cmap, colorbar, avg_method


def _plot_surfs(data, space=None, hemi=None, surf_mesh=None,
                bg_map=None, rois=False, cmap='default',
                bg_on_data=.25, darkness=None,  avg_method='default',
                wspace=-.35, hspace=-.1, alpha=1,
                threshold='auto', symmetric_cbar='auto', 
                colorbar='default', _print=None, **kwargs):

    # Process default surface and plotting values
    data, ref, symmetric_cbar, threshold, cmap, colorbar, avg_method =\
        _prep_auto_defaults(data, space, hemi, rois,
                            symmetric_cbar, threshold,
                            cmap, colorbar, avg_method, _print=_print)
 
    # If user-passed - update params
    surf_mesh, bg_map, darkness =\
        _proc_ref_arg_defaults(ref, surf_mesh, bg_map, darkness)

    # Both hemi's passed case.
    if 'lh' in data and 'rh' in data:
        return plot_surf_collage(data=[data['lh'], data['rh']],
                                 ref=ref,
                                 surf_mesh=surf_mesh,
                                 bg_map=bg_map,
                                 cmap=cmap,
                                 avg_method=avg_method,
                                 threshold=threshold,
                                 symmetric_cbar=symmetric_cbar,
                                 alpha=alpha,
                                 bg_on_data=bg_on_data,
                                 darkness=darkness,
                                 wspace=wspace, hspace=hspace,
                                 colorbar=colorbar, **kwargs)

    # Just lh/ rh cases
    if 'lh' in data:
        hemi_data, hemi = data['lh'], 'left'
    if 'rh' in data:
        hemi_data, hemi = data['rh'], 'right'
    
    # Plot
    return plot_surf_hemi(data=hemi_data,
                          ref=ref,
                          hemi=hemi,
                          surf_mesh=surf_mesh,
                          bg_map=bg_map,
                          cmap=cmap,
                          avg_method=avg_method,
                          threshold=threshold,
                          symmetric_cbar=symmetric_cbar,
                          alpha=alpha,
                          bg_on_data=bg_on_data,
                          darkness=darkness,
                          colorbar=colorbar,
                          **kwargs)

def _sort_kwargs(kwargs):

    # Surface specific
    surf_args = ['alpha', 'bg_map',
                 'surf_mesh', 'darkness',
                 'view', 'dist', 'b_sz']
    surf_params = {key: kwargs[key] for key in surf_args if key in kwargs}
    
    # Special cases
    if 'surf_hspace' in kwargs:
        surf_params['hspace'] = kwargs['surf_hspace']

    # If also passed directly
    # Give priority to any if passed in surf_params vs. kwargs
    if 'surf_params' in kwargs:
        surf_params = {**surf_params, **kwargs['surf_params']}
    
    # Volume specific
    vol_args = ['resampling_interpolation', 'plot_abs',
                'black_bg', 'display_mode', 'annotate', 'cut_coords',
                'bg_img', 'draw_cross', 'dim', 'view_type', 'linewidths']
    vol_params = {key: kwargs[key] for key in vol_args if key in kwargs}

    # Special case
    if 'vol_alpha' in kwargs:
        vol_params['alpha'] = kwargs['vol_alpha']

    # If also passed directly
    # Give priority to any in passed vol params
    if 'vol_params' in kwargs:
        vol_params = {**vol_params, **kwargs['vol_params']}
    
    # For color bar params
    colorbar_params = _sort_colorbar_kwargs(colorbar_params=None, **kwargs)

    return surf_params, vol_params, colorbar_params

def _sort_colorbar_kwargs(colorbar_params=None, **kwargs):


    # Handle if not passed
    if colorbar_params is None:
        if 'colorbar_params' in kwargs:
            colorbar_params = kwargs.pop('colorbar_params')
        else:
            colorbar_params = {}

    # Colorbar specific args
    colorbar_args = ['fraction', 'shrink', 'aspect',
                     'pad', 'anchor', 'format', 'cbar_fig_ratio']
  
    # Pop these arguments
    kwargs_colorbar_params = {key: kwargs.pop(key) for key in colorbar_args if key in kwargs}

    # Combine
    colorbar_params = {**colorbar_params, **kwargs_colorbar_params}

    return colorbar_params

def _proc_cmap(cmap, rois, symmetric_cbar, flat_data):
    
    # Keep user passed if not user passed
    if not (cmap == 'default' or cmap is None or cmap == 'auto'):
        return cmap
    
    # If plotting rois
    if rois:
        return 'prism'
    
    # If not symmetric, then just do Reds or Blues
    if symmetric_cbar is False:

        # If max value is 0 or less, use Blues
        if np.nanmax(flat_data)  <= 0:
            return 'Blues_r'

        # Otherwise, reds
        return 'Reds'

    # Last case is symmetric cbar
    return _trunc_cmap(plt.get_cmap('cold_white_hot'), minval=0.05, maxval=0.95)


def _plot_surfs_vol(data, space=None, hemi=None,
                    rois=False, cmap='default',
                    vol_plot_type='glass',
                    surf_mesh=None, bg_map=None,
                    colorbar='default', symmetric_cbar='auto',
                    darkness=None, vmin=None, vmax=None,
                    cbar_vmin=None, cbar_vmax=None,
                    figure=None, subplot_spec=None,
                    figsize='default', title=None,
                    title_sz=18, title_y=None,
                    hspace='default', wspace=-.2, alpha=1,
                    avg_method='default', threshold='auto',
                    surf_to_vol_ratio='default', surf_wspace='default',
                    bg_on_data=.25, _print=None, **kwargs):

    # Process default surface and plotting values
    data, ref, symmetric_cbar, threshold, cmap, colorbar, avg_method =\
        _prep_auto_defaults(data, space, hemi, rois,
                            symmetric_cbar, threshold,
                            cmap, colorbar, avg_method, _print=_print)
    
    # If user-passed
    surf_mesh, bg_map, darkness = _proc_ref_arg_defaults(ref, surf_mesh, bg_map, darkness)

    # Set default anchor value
    if 'anchor' not in kwargs:
        kwargs['anchor'] = (0, .6)

    # Set specific plot settings if glass or not
    # due to differences in sizes
    if vol_plot_type == 'glass':

        if surf_to_vol_ratio == 'default':
            surf_to_vol_ratio = 1.1
        if hspace == 'default':
            hspace = -.2
        if surf_wspace == 'default':
            surf_wspace = -.125

    # Roi's / plot stat volume case
    else:
        if surf_to_vol_ratio == 'default':
            surf_to_vol_ratio = .9
        if hspace == 'default':
            hspace = -.25
        if surf_wspace == 'default':
            surf_wspace = -.2
    
    # Set default figsize based on if colorbar
    if figsize == 'default':

        if colorbar:
            figsize = (12, 12)
        else:
            figsize = (9, 12)

    # Sort kwargs into categories
    surf_params, vol_params, colorbar_params = _sort_kwargs(kwargs)

    # If plotting rois, force values to plot roi specific
    if rois:
        vol_plot_type = 'roi'

    # Add surface specific to surf_params dict
    surf_params['surf_mesh'], surf_params['bg_map'] = surf_mesh, bg_map
    surf_params['darkness'], surf_params['alpha'] = darkness, alpha
    surf_params['ref'], surf_params['avg_method'] = ref, avg_method
    surf_params['wspace'], surf_params['bg_on_data'] = surf_wspace, bg_on_data
    
    # Pass arguments to plot surf vol, and return
    return plot_surf_vol_collage(surf=[data['lh'], data['rh']],
                                 vol=data['vol'], vol_plot_type=vol_plot_type,
                                 cmap=cmap, threshold=threshold, vmin=vmin, vmax=vmax,
                                 cbar_vmin=cbar_vmin, cbar_vmax=cbar_vmax,
                                 figure=figure, subplot_spec=subplot_spec,
                                 figsize=figsize, title=title, title_sz=title_sz, title_y=title_y,
                                 colorbar=colorbar, symmetric_cbar=symmetric_cbar,
                                 hspace=hspace, wspace=wspace, colorbar_params=colorbar_params,
                                 surf_params=surf_params, vol_params=vol_params,
                                 surf_to_vol_ratio=surf_to_vol_ratio, _print=_print)

def plot_volume(vol,
                vol_plot_type='glass',
                cmap=None,
                colorbar='default',
                symmetric_cbar='auto',
                threshold='auto',
                vmax=None,
                figure=None,
                axes=None,
                **kwargs):

    # Base stat map cases
    rois = False
    if vol_plot_type == 'glass':
        vol_plot_func = plot_glass_brain
    elif vol_plot_type == 'stat':
        vol_plot_func = plot_stat_map
    
    # ROI case
    elif vol_plot_type == 'roi':
        vol_plot_func = plot_rois
        rois = True

    # Process defaults
    symmetric_cbar, threshold, cmap, colorbar, _ = _prep_base_auto_defaults(vol, rois,
                                                                            symmetric_cbar, threshold,
                                                                            cmap, colorbar, avg_method=None)

    # Call vol plot function
    vol_plot_func(vol, figure=figure,  axes=axes,
                  cmap=cmap, threshold=threshold, 
                  symmetric_cbar=symmetric_cbar,
                  vmax=vmax, colorbar=colorbar, **kwargs)
    
    # Set if used by upper level functions to return
    vol_smfs = [np.array([np.nanmin(vol.get_fdata()), np.nanmax(vol.get_fdata())])]

    return figure, axes, vol_smfs


def _setup_auto_plot(data, space=None, hemi=None, verbose=0, **kwargs):

    # Get verbose object
    _print = _get_print(verbose=verbose)
    
    # Load / perform initial auto-detection of data
    data, _ = _load_data_and_ref(data, space=space, hemi=hemi, _print=_print)

    # If includes surface data -
    # Get the unique values
    if 'lh' in data:
        unique_vals = np.unique(data['lh'])
    elif 'rh' in data:
        unique_vals = np.unique(data['rh'])
    # If no surface data present, subcort case
    else:
        unique_vals = np.unique(data['vol'].get_fdata())

    # Check for passed rois value
    if 'rois' in kwargs:
        rois = kwargs.pop('rois')
    else:
        rois = None

    # If all of the data points are interger-like, assume
    # we are plotting a parcellation and not stat data
    is_ints = all([float(u).is_integer() for u in unique_vals])

    # If either passed rois True, or is all is ints and not rois is False
    if rois is True or (is_ints and rois is not False):
        
        # Just surface or surfaces case
        if 'vol' not in data:
            return data, _plot_surfs, {'space': space,
                                       'hemi': hemi,
                                       'rois': True,
                                       '_print': _print,
                                       **kwargs}
        
        # Just volume case
        elif 'lh' not in data and 'rh' not in data:
           
            # Override vol_plot_type - force roi
            return data['vol'], plot_volume, {'vol_plot_type': 'roi', **kwargs}
            
        # Last case is vol / surf collage

        return data, _plot_surfs_vol, {'space': None,
                                       'hemi': None,
                                       'rois': True,
                                       '_print': _print,
                                       **kwargs}
    
    # Otherwise - stat values / not ROIs

    # Just surface or surfaces case
    if 'vol' not in data:
        return data, _plot_surfs, {'space': space,
                                   'hemi': hemi,
                                   'rois': False,
                                   '_print': _print,
                                   **kwargs}

    # Just volume case
    elif 'lh' not in data and 'rh' not in data:
        return data['vol'], plot_volume, kwargs

    # Last case is vol / surf collage
    return data, _plot_surfs_vol, {'space': None,
                                   'hemi': None,
                                   'rois': False,
                                   '_print': _print,
                                   **kwargs}

def _check_subplot_args(**kwargs):

    # If already a subplot spec passed, then just check for figure, then return.
    if 'subplot_spec' in kwargs and kwargs['subplot_spec'] is not None:
        
        # If figure is passed and not None, we don't need to set
        if 'figure' in kwargs and kwargs['figure'] is not None:
            pass
        else:
            kwargs['figure'] = plt.gcf()
        
        # Then return
        return kwargs

    # If ax is passed, but subplot_spec not, then change to subplot version
    if 'ax' in kwargs:

        # Remove from kwargs
        ax = kwargs.pop('ax')

        # If not None, continue w/ figure check / convert to subplot
        if ax is not None:

            # Set off, for plotting
            ax.set_axis_off()

            # If figure is passed and not None, we don't need to set
            if 'figure' in kwargs and kwargs['figure'] is not None:
                pass
            else:
                kwargs['figure'] = plt.gcf()

            # Convert ax to subplot spec
            ax_as_subplot_spec = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=ax)

            # Add in as subplot spec, replacing passed ax
            kwargs['subplot_spec'] = ax_as_subplot_spec[0, 0]

    return kwargs

def plot(data, space=None, hemi=None, verbose=0, returns=False, **kwargs):
    '''The most automated magic plotting function avaliable,
    used to plot a wide range of neuroimaging data (volumes / surfs).

    Parameters
    -----------
    data : str, array, dict, ect...
        The data in which to plot, either statistical map or parcellaction,
        as broadly either a single surface, collage of surfaces, a single
        volume, or collage with surfaces and volume.
    
        Data can be passed in many ways:

        - As array-like input representing a single surface hemisphere,
          concatenated hemispheres, concatenated hemispheres + flattened sub-cortical
          values, or just flattened volumetric values.

        - As a str, representing the file location with saved values
          for any of the above array options. A large number of different
          file formats are accepted.

        - As a dictionary. If passed as a dictionary, this
          allows the user most control over how data is specified.
          Dictionary keys must be one or more of 'lh', 'rh' and 'vol',
          where values can be either array-like, file locations or nibabel
          objects. Note that if you are trying to plot data in native
          subject space, it must be passed in this dictionary style, along
          with, if a surface, a proper value for surf_mesh.

        - As a list or array-like of length 3, indicating internally
          that the passed values represent in the first index the file location
          or array values of a left hemisphere surface, and in the second index,
          the right hemisphere surface.

        - As a list or array-like of length 3, indicating internally
          that the first two passed index are left and right hemisphere
          surface data, and the last is subcortical / volumetric data, all to
          be plotted together in a collage.

        - As a :class:`Nifti1Image<nibabel.nifti1.Nifti1Image>` to be plotted volumetrically.

    space : None or str, optional
        This argument defines the "space" in which
        surface data is plotted. If left as default (None),
        the space in which the data to plot is
        in will be automatically inferred,
        otherwise this parameter can be manually set,
        overriding the automatically detected choice.
        Current supported options are:

        - 'fsaverage' : freesurfer average standard space.

        - 'fsaverage5' : freesurfer5 downsampled average standard space.

        - 'fsaverage4' : freesurfer3 downsampled average standard space.

        - 'fsaverage3' : freesurfer3 downsampled average standard space.

        - '41k_fsaverage' : Matched to CIVET 41k vertex freesurfer space.

        - '4k_fs_LR' :  HCP standard 4k vertex space. 

        - '8k_fs_LR' :  HCP standard 8k vertex space. 

        - '32k_fs_LR' : HCP standard 32k vertex space. 

        - '59k_fs_LR' : HCP standard 59k vertex space. 

        - '164k_fs_LR' : HCP standard 164k vertex space. Note, while this space has the same number of vertex as fsaverage space, it is not the same.

        - 'civet' : Standard space with 41k vertex used by CIVET.

        - 'native' : This is set if the passed number of values don't correspond to any of the saved defaults, and refers to that the space to plot is in some non-standard space.

        ::

            default = None

    hemi : None, 'lh' or 'rh', optional
        This parameter is only used when plotting
        any surface data. Further, this parameter
        is only relevant the surface data being plotted represents
        a single hemispheres data.
        
        If left as default, None,
        then this option will be automatically set to plot
        left hemisphere data.

        ::

            default = None

    verbose : int, optional
        Plotting includes a many, many of parameters,
        which is why this function is helpful as it automates a
        large number of choices. That said, it can helpful
        to be aware of what choices are being made, which is
        what this verbose parameter controls. The following
        levels are made avaliable:

        - <0 : Nothing, not even warnings are shown.
        - 0 : Only warnings are shown.
        - 1 : Information on which steps are automatically decided are shown.
        - >1 : Automatic choices are shown as well as additional helper text.

    returns : bool, optional
        If the figure, axis / grid and smfs should be returned from the base function
        calls or not.

        ::

            default = False

    kwargs : keyword arguments
        There are number of different plotting specific arguments
        that can optionally be modified. The default values for these arguments
        may change also depending on different auto-detected settings, i.e.,
        if plotting a single surface vs. surface collage, or if plotting
        a volumetric image. The remaining parameters below are these kwargs.

        - **surf_mesh** : str or array
        
          A str indicator specifying a valid surface mesh,
          with respect to the current space, or a two dimensional array
          with information about the coordinates and vertex-faces
          in which to plot the data on. The default for
          freesurfer spaces is 'inflated' and for fs_LR spaces
          is 'very_inflated'. If you wish to plot surface
          data in native subject space, then this argument must be
          supplied with a loaded 2D array / valid mesh, and data
          must be passed in dictionary style.

          This parameter is only relevant when plotting surfaces.

        - **bg_map** : str or 2D array
        
          This argument specifies a background map to be used when plotting. This
          should be passed as either a str indicator specifying a valid file
          within the current surface space, or as a valid array of values,
          again, with respect to the current space. This map is plotted in greyscale
          underneath the data points, often used for realistic shading.
          The default for when plotting is 'sulc'.

        - **cmap** : str or :class:`matplotlib.colors.Colormap`
          
          This should be pass as an instance
          of :class:`matplotlib.colors.Colormap` or str
          representing the name of a matplotlib colormap. This will be the color map
          in which the values are plotted according to. When plotting surface
          parcellations, the default cmaps are 'prism' if plotting rois, 'Reds'
          if plotting not symmetric statistical maps, and 'cold_white_hot' if plotting
          symmetric statistical maps (e.g., values above and below 0). In case of
          not sym plotting, and plotting  negative numbers, will switch from 'Reds'
          to 'Blues_r'.

        - **vol_plot_type** : {'glass', 'stat', 'roi'}
        
          This parameter control the type of volumetric plot in which to generate,
          with valid values as one of 'glass', 'stat' and 'roi'. By default,
          if detected to be plotting a parcellation, the value 'roi' is used.
          Otherwise, the default volumetric plotting type is 'glass'.
          The corresponding back-end nilearn functions used are:

          - 'glass': :func:`nilearn.plotting.plot_glass_brain`
          - 'stat': :func:`nilearn.plotting.plot_stat_map`
          - 'roi': :func:`nilearn.plotting.plot_roi`

          *This parameter is only used when plotting volumetric data.*

        - **bg_on_data** : bool or float
          If True, and a bg_map is specified,
          the data to plot is multiplied by the background image, 
          so that e.g. sulcal depth is visible beneath the surf_data.

          If passed as a float value, then that will trigger multiplying
          the data to plot by the background image times the passed bg_on_data
          value. So for example passing True is equivalent to passing 1 and passing
          False is the same as passing 0.
          
          This allows for fine tuning how much the background image
          is melded with the data shown. The default value is .25.

          *This parameter is only used when plotting surfaces.*

        - **view** : str
        
          If plotting a single surface hemisphere, this
          parameter must be one of:

            {'lateral', 'medial', 'dorsal', 'ventral', 'anterior', 'posterior'}
          
          Where 'lateral' is the default view if none is set.

          If instead plotting a collage of surface views, then
          valid view parameters are one of:

            {'standard', 'front back', 'front', 'back'}
          
          which correspond to different preset collections of surface views.
          The default if not specified in this case is 'standard'.
        
          Note: If plotting a collage of surface and volumetric views,
          it is reccomended to keep the default collage 'standard' view,
          as the other views have not yet been properly formatted yet.

        - **colorbar** : bool

          This option dictates if a colorbar should be added
          along with the requested plot, regardless of type.
          By default a colorbar will be added (True) unless
          a parcellation is being plotted, in which the default is
          to not add a colorbar (False).

          There are a number of extra more detailed colorbar style
          kwargs, used in certain cases, with certain plots to
          customize the look and size of the colorbar. See below.

        - **rois** : bool
          
          If passed, can force plot either is rois case or stat map case.
          Ussually the auto settings will be fine, but this exists if need to override
          without overriding all of the other little settings. Default is just automatically
          detected based on percent of unique values being plotted.

        - **avg_method** : 'mean', 'median', 'min', 'max' or  custom function

          This option defines how to average vertex values to
          derive face values, and is only used when plotting surfaces.

          - 'mean': results in smooth boundaries

          - 'median': results in sharp boundaries

          - 'min' or 'max': for sparse matrices

          - custom function: You can also pass a custom function

          If plotting roi's the default will be 'median'. Otherwise, the default
          will be set based on the ratio between the number of unique values
          to data points. So if more than 5% of data points are unique, 'mean'
          will be used, otherwise 'median' will be used. This is to try and accurately detect
          when parcellations are being plotted.

        - **alpha** : float or 'auto'

          This parameter is only used when plotting surface data. It
          refers to the alpha level of the mesh. By default, this
          is left as 'auto', which will default to .5 if no bg_map
          and 1 if a bg_map is passed. This parameter is just multiplied by
          the face colors before plotting.

          Note: this keyword is used for plotting surfaces, though alpha is
          also used as a parameter for plotting the volumetric glass brain.
          To access this parameter for the volumetric function, see the
          `vol_alpha` parameter.

        - **vol_alpha** : float between 0 and 1
        
          This controls the alpha transparency only when plotting
          volumetric data according to the vol_plot_type='glass' option.
          The default value is 0.7.
        
        - **darkness** : float between 0 and 1

          This parameter specified the darkness of the background image,
          when plotting a single or multiple surfaces. Where a value of 1
          indicates that the original values are used, and .5 would mean the
          values are halved. If plotting in an fsaverage space the default value is 1,
          if plotting in an fs_LR space, the default value is .5. 

        - **symmetric_cbar** : bool or 'aut'
        
            Specifies whether the colorbar should range from -vmax to vmax or from vmin to vmax.
            Setting to 'auto' will select the latter if the range of
            the whole image is either positive or negative. There are some
            other automatic cases, but in general this parameter
            can just be used to override any automatic choices.

        - **threshold** : float, None or 'auto'

          If None, the image is not thresholded. If a float,
          or other numberic value, that value will be used to threshold
          the image. This threshold is treated as an absolute value in
          the case of a symmetric_cbar. If 'auto' is passed,
          it will attempt to automatically set the threshold to
          a reasonable value. 

          The default is typically 'auto', but in some cases changes,
          for example if plotting a parcellation the default value will be
          None.s

        - **figsize** : 'default' or (int, int)

          This parameter is used when axes are not passed, and
          a new figure is being generated. It represents
          the size of the underlying matplotlib figure to plot on.
          The default value for this function varies based on what type of
          plot is being made (e.g., not the same for surface collage vs. single surface).

        - **wspace** : float

          This parameter refers to the width spacing between items at the top
          level of whatever collage is being plotted. So, if plotting only a single
          surface or volume, this parameter will be ignored. Otherwise,
          for example if plotting a collage of 4 surface views, then this
          parameter will control the amount of horizontal space between each
          brain view. Instead, if plotting a combined collage of surface views
          and volumetric views, then the top level of the collage that this parameter
          controls the spacing of is the set of all surface views, all volumetric views
          and colorbar. In this case, to override values between for example just the
          surface views, you would have to use special extra keyword 'surf_wspace',
          which let's you set both parameters if desired.

          The default values hover around 0, and vary a great deal based on the type
          of plot and requested view.

        - **hspace** : float

          This parameter refers to the height spacing between items at the top
          level of whatever collage is being plotted. So, if plotting only a single
          surface or volume, this parameter will be ignored. Otherwise,
          for example if plotting a collage of 4 surface views, then this
          parameter will control the amount of vertical space between each
          brain view. Instead, if plotting a combined collage of surface views
          and volumetric views, then the top level of the collage that this parameter
          controls the spacing of is the set of all surface views, all volumetric views
          and colorbar. In this case, to override values between for example just the
          surface views, you would have to use special extra keyword 'surf_hspace',
          which let's you set both parameters if desired.

          The default values hover around 0, and vary a great deal based on the type
          of plot and requested view.

        - **surf_wspace** : float
          
          See 'wspace'. This parameter is only used when there
          are multiple levels of collage and further control is still
          required for specifically the nested collage of surface plots.

        - **surf_hspace** : float
          
          See 'hspace'. This parameter is only used when there
          are multiple levels of collage and further control is still
          required for specifically the nested collage of surface plots.

        - **surf_to_vol_ratio** : float or 'default'

          In the case of plotting a collage of both surfaces and
          volumetric data, this parameter can optionally be set.
          It defines the size ratio between the grouping of surface collages
          and the volumetric plots. If left as 'default', will
          be set to either .9 in base plotting stat volume case
          and 1.1 when plotting glass brain.

    Notes
    -----------
    The creation of carefully constructed multi-figures can be a little tricky,
    which is why it is helpful that so many of the default values have been set. That
    said, it is likely the interested user could do better than the defaults with
    say for example plotting a multi-figure surface + volume collage. In this instance,
    if they wanted to tweak the positioning of the different sub figures relative to
    each other, they could make use of the following parameters:

    - hspace
    - wspace
    - surf_hspace
    - surf_wspace
    - surf_to_vol_ratio
    - cbar_fig_ratio
    - figsize

    '''

    # Some logic for handling if this is going to be a subplot
    kwargs = _check_subplot_args(**kwargs)
    
    # Setup
    data, func, args = _setup_auto_plot(data, space=space, hemi=hemi,
                                        verbose=verbose, **kwargs)

    # Call func
    r = func(data, **args)

    # Optionally return Figure + whatever returned by sub calls
    if returns:
        return r

    return None


def _get_is_roi(args):
    
    # Vol case
    if 'vol_plot_type' in args:
        if args['vol_plot_type'] == 'roi':
            return True
    
    # Surf case
    elif 'rois' in args:
        if args['rois']:
            return True
        
    return False

def _get_data_func_types(data):
    
    data_func_types = {}
    data_as_list = []
    is_rois = []
    
    for key in list(data):
        loaded, func, args = _setup_auto_plot(data[key], space=None,
                                              hemi=None, verbose=0)
        data_func_types[key] = func.__name__
        data_as_list.append(loaded)
        is_rois.append(_get_is_roi(args))

    return data_func_types, data_as_list, is_rois


def _add_grid_wh(sz, layout_params, max_cols=2):
    
    # Return single row, w/ length sz
    if sz < max_cols:
        ws = [1 for _ in range(sz)]
        hs = [1]
    
    # Otherwise calc as grid, w/ leave empty any missing
    else:
        n_rows = (sz+1) // max_cols
        ws = [1 for _ in range(max_cols)]
        hs = [1 for _ in range(n_rows)]
        
    # Add to layout params
    layout_params['widths'] = ws
    layout_params['heights'] = hs
    
    # Return n_rows, n_cols for conv.
    return len(ws), len(hs)

def _get_def_layout_params(unique_func_types, data_keys, colorbar, colorbar_params):

    # Init dict of layout params
    layout_params = {}

    # All same cases
    if len(unique_func_types) == 1:
        
        # If all same, then get single type
        single_type = unique_func_types[0]

        # TODO change max cols maybe based on single_type
        
        # Gen base grid based on number of things to plot
        n_cols, n_rows = _add_grid_wh(len(data_keys), layout_params, max_cols=2)

        # Collage of just surfaces
        if single_type == '_plot_surfs':
        
            # Base colorbar vs. no cbar settings, first default None case
            layout_params['title_y'] = None
            layout_params['sub_title_y'] = .9 - ((n_rows-1)/100)
            layout_params['title_sz'] = 27
            
            # Colorbar case
            if colorbar:
                layout_params['title_y'] = .95
                layout_params['sub_title_y'] = .83
                layout_params['title_sz'] = 24
            
            # Set fig size
            layout_params['figsize'] = (10*n_cols, 10*n_rows)
            
            # Adjust title sz
            layout_params['title_sz'] += (n_rows-1) * 3
            
            # Other layout settings
            layout_params['wspace'] = .15
            layout_params['sub_wspace'] = .1
            
            if colorbar:
                if n_rows == 1:
                    layout_params['hspace'] = 0
                    layout_params['sub_hspace'] = -.5
                elif n_rows == 2:
                    layout_params['hspace'] = -.35
                    layout_params['sub_hspace'] = -.58
                else:
                    raise RuntimeError()
           
            else:
                if n_rows == 1:
                    layout_params['hspace'] = 0
                    layout_params['sub_hspace'] = -.31
                elif n_rows == 2:
                    layout_params['hspace'] = -.185
                    layout_params['sub_hspace'] = -.390
                elif n_rows > 2:
                    layout_params['hspace'] = -.185 - ((n_rows - 2)  * .025)
                    layout_params['sub_hspace'] = -.390 - ((n_rows - 2) * .02)

            
            # Only set fraction if one row
            if n_rows == 1:
                colorbar_params['fraction'] = 1

        # Surf and vol case
        elif single_type == '_plot_surfs_vol':
            
            layout_params['title_y'] = .95
            layout_params['sub_title_y'] = .83
            layout_params['title_sz'] = 24

    return layout_params

def meta_collage(data, title=None, sub_titles=True,
                 colorbar='default', verbose=0, vmin=None,
                 vmax=None, colorbar_params=None, threshold='auto',
                 symmetric_cbar='auto', cmap='default', 
                 avg_method='default', sub_kwargs=None, **kwargs):
    
    # Check explicit sub kwargs
    if sub_kwargs is None:
        sub_kwargs = {}
    
    # Check initial data
    data_func_types, data_as_list, is_rois = _get_data_func_types(data)
    unique_func_types = list(set(data_func_types.values()))
    data_keys = list(data)
    
    # Catch error
    if len(set(is_rois)) > 1 and colorbar:
        raise RuntimeError('Cant plot mix of rois and stat maps with colorbar.')
    
    # Passes True if all True
    rois = all(is_rois)
    
    # Collapse data
    flat_data = _collapse_data(data_as_list)
    
    # Process some base auto args
    symmetric_cbar, threshold, cmap, colorbar, avg_method =\
        _prep_base_auto_defaults(flat_data, rois,
                                 symmetric_cbar, threshold,
                                 cmap, colorbar, avg_method)

    
    # Process vmin / vmax from data and passed args
    vmin, vmax = _proc_vs(data_as_list, vmin=vmin, vmax=vmax,
                          symmetric_cbar=symmetric_cbar)
    
    # Proc default / add params
    colorbar_params = _sort_colorbar_kwargs(colorbar_params, **kwargs)

    # Get default layout params
    layout_params = _get_def_layout_params(unique_func_types, data_keys, colorbar, colorbar_params)
    
    # Need to check if any user over rides in layout params
    for key in layout_params:
        
        # If user passed, favor that
        if key in kwargs:
            layout_params[key] = kwargs.pop(key)
            
    # Fill in sub kwargs
    def_sub_kwargs = {}
    def_sub_kwargs['wspace'] = layout_params.pop('sub_wspace')
    def_sub_kwargs['hspace'] = layout_params.pop('sub_hspace')
    def_sub_kwargs['title_y'] = layout_params.pop('sub_title_y')
    
    if 'sub_title_sz' in kwargs:
        def_sub_kwargs['title_sz'] = kwargs.pop('sub_title_sz')
        
    # Idea here is to give priority to anything passed
    # explicitly with sub_kwargs, then kwargs, then defaults
    sub_kwargs = {**kwargs, **def_sub_kwargs, **sub_kwargs}
                
    # Setup base collage settings / grid spec
    figure, grid, colorbar_ax = _setup_fig_axes(get_grid=True,
                                                title=title,
                                                colorbar=colorbar,
                                                colorbar_params=colorbar_params,
                                                **layout_params)
    
 
    # Add sub plots
    smfs = []
    for i in range(len(layout_params['heights'])):
        for j in range(len(layout_params['widths'])):
            key = data_keys[i+j]
            
            # Optional sub title
            if isinstance(sub_titles, bool):
                if sub_titles:
                    sub_title = str(key)
                else:
                    sub_title = None

            # Otherwise if list
            elif isinstance(sub_titles, list):
                sub_title = sub_titles[i+j]
            
            # Or dict
            elif isinstance(sub_titles, dict):
                sub_title = sub_titles[key]
            
            # Last case is just None
            else:
                sub_title = None

            # Call plot
            _, _, smfz = plot(data[key],
                              figure=figure,
                              subplot_spec=grid[i, j],
                              colorbar=False,
                              title=sub_title,
                              symmetric_cbar=symmetric_cbar,
                              threshold=threshold,
                              cmap=cmap, vmin=vmin, vmax=vmax,
                              avg_method=avg_method, verbose=verbose,
                              returns=True,
                              **sub_kwargs)
            
            
            smfs.append(smfz)
            
    # Set to concat
    smfs = np.concatenate(smfs)
    
    # Add collage color bar
    if colorbar:
        add_collage_colorbar(figure=figure,
                             ax=colorbar_ax,
                             smfs=smfs,
                             vmin=vmin, vmax=vmax,
                             multicollage=True,
                             colorbar_params=colorbar_params,
                             threshold=threshold,
                             cmap=cmap,
                             **kwargs)



import os
import nibabel as nib
from ...loading import load
from ...transform import add_surface_medial_walls, SurfLabels
from ..plot import plot
import numpy as np
import matplotlib.pyplot as plt

nib.imageglobals.logger.level = 40
file_dr = os.path.dirname(os.path.realpath(__file__))

def test_plotting_and_transforms_case():
    
    # Load ROI labels
    glasser_labels = np.random.randint(0, 10, size=64984)
    
    # Load without transforms make sure works
    data = np.random.random(size=64984)

    # Test generic smart plot function for
    # just making sure works w/o errors
    plot(data, colorbar=True)

    # Test Extract ROIs
    surf_labels = SurfLabels(glasser_labels, vectorize=False)
    rois = surf_labels.fit_transform(data)
    assert rois.shape[0] == 9

    # Test plot again, and also reverse transform function
    plot(surf_labels.inverse_transform(rois), colorbar=True)

    plt.close()


import numpy as np
import warnings
from scipy.linalg import lstsq
from sklearn.base import BaseEstimator, TransformerMixin
from ..loading import load
import nibabel as nib
import pandas as pd

class SurfLabels(BaseEstimator, TransformerMixin):
    '''Extract signals from non-overlapping labels for surface data.

    This class functions similar to :class:`NiftiLabelsMasker<nilearn.input_data.NiftiLabelsMasker>`,
    except is designed to work with surface data. 

    Parameters
    -----------
    labels : str or array-like
        This should represent an array, of the same size as the data
        dimension, as a mask with unique integer values for each ROI.
        You can also pass a str location in which to load in this array.
        Anything accepted by :func:`load<neurotools.loading.load>` is acceptable here.

    background_label : int, array-like of int, optional
        This parameter determines which label, if any,
        in the corresponding passed labels,
        should be treated as 'background' and therefore no ROI
        calculated for that value or values.
        You may pass either a single interger
        value, an array-like of integer values.

        If not background label is desired, just pass
        a label which doesn't exist in any of the data,
        e.g., -2.

        ::

            default = 0

    mask : None, str or array-like, optional
        This parameter allows you to optional pass a mask of values in
        which to not calculate ROI values for.
        This can be passed as a str or
        array-like of values (just like labels),
        and should be comprised of
        a boolean array (or 1's and 0's),
        where a value of 1 means that value
        will be ignored (set to background label)
        should be kept, and a value of 0,
        for that value should be masked away.
        This array should have the same length as the passed `labels`.

        ::

            default = None

    strategy: specific str, custom_func, optional
        This parameter dictates the function to be applied
        to each data's ROI's
        individually, e.g., mean to calculate the mean by ROI.

        If a str is passed, it must correspond to one of the below preset
        options:

        - 'mean'
            Calculate the mean with :func:`numpy.mean`

        - 'sum'
            Calculate the sum with :func:`numpy.sum`

        - 'min' or 'minimum
            Calculate the min value with :func:`numpy.ndarray.min`

        - 'max' or 'maximum
            Calculate the max value with :func:`numpy.ndarray.max`

        - 'std' or 'standard_deviation'
            Calculate the standard deviation with :func:`numpy.std`

        - 'var' or  'variance'
            Calculate the variance with :func:`numpy.var`

        If a custom function is passed, it must accept two arguments,
        custom_func(X_i, axis=data_dim), X_i, where X_i is a subjects data
        array where that subjects data corresponds to
        labels == some class i, and can potentially
        be either a 1D array or 2D array, and an axis argument
        to specify which axis is
        the data dimension (e.g., if calculating for a time-series
        [n_timepoints, data_dim], then data_dim = 1,
        if calculating for say stacked contrasts where
        [data_dim, n_contrasts], data_dim = 0, and lastly for a 1D
        array, data_dim is also 0.

        ::

            default = 'mean'

    vectorize : bool, optional
        If the returned array should be flattened to 1D. E.g., if the
        last step in a set of loader steps this should be True, if before
        a different step it may make sense to set to False.

        ::

            default = True

    See Also
    -----------
    SurfMaps : For extracting non-static / probabilistic parcellations.
    nilearn.input_data.NiftiLabelsMasker : For working with volumetric data.

    '''

    def __init__(self, labels,
                 background_label=0,
                 mask=None,
                 strategy='mean',
                 vectorize=True):

        self.labels = labels
        self.background_label = background_label
        self.mask = mask
        self.strategy = strategy
        self.vectorize = vectorize

    def _base_fit(self):

        # Load mask if any
        self.mask_ = load(self.mask)

        # Load labels
        self.labels_ = load(self.labels)

        if len(self.labels_.shape) > 1:
            raise RuntimeError('The passed labels array must be flat, ' +
                               ' i.e., a single dimension')

        # Mask labels if mask
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != len(self.labels_):
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the length of labels!')
            self.labels_[self.mask_.astype(bool)] = self.background_label

        # Proc self.background_label, if int turn to np array
        if isinstance(self.background_label, int):
            self.background_label_ = np.array([self.background_label])

        # Otherwise, if already list-like, just cast to np array
        else:
            self.background_label_ = np.array(self.background_label)

        # Set the _non_bkg_unique as the valid labels to get ROIs for
        self.non_bkg_unique_ = np.setdiff1d(np.unique(self.labels_),
                                            self.background_label_)

        # Proc strategy if need be
        strats = {'mean': np.mean,
                  'median': np.median,
                  'sum': np.sum,
                  'minimum': np.min,
                  'min': np.min,
                  'maximum': np.max,
                  'max': np.max,
                  'standard_deviation': np.std,
                  'std': np.std,
                  'variance': np.var,
                  'var': np.var}

        if self.strategy in strats:
            self.strategy_ = strats[self.strategy]
        else:
            self.strategy_ = self.strategy

    def fit(self, X, y=None):
        '''Fit this object according
        the passed subjects data, X.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            matches the shape of the passed
            labels (if 1D, the shapes must match)
            and the other, optionally, represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None
        '''

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')
        
        # Perform pieces of fit not based on passed data
        self._base_fit()
        
        # Make sure labels matches X shape
        if len(self.labels_) not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform this object.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''
        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "labels_"):
            raise ValueError('It seems that SurfLabels has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):
        '''Transform this the passed data.

        If X has the both the same dimension's, raise warning.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''

        if not hasattr(self, "labels_"):
            self._base_fit()

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length' +
                          ' in each dimension, ' +
                          'Assuming that axis=0 is the data dimension' +
                          ' w/ vertex values')

        if len(X.shape) > 2:
            raise RuntimeError('The shape of X can be at most 2 dimensions.')

        # The data dimension is just the dimension with
        # the same len as the label
        self.data_dim_ = X.shape.index(len(self.labels_))
        self.X_shape_ = X.shape

        # Get the ROI value for each label
        X_trans = []
        for i in self.non_bkg_unique_:

            if self.data_dim_ == 0:
                X_i = X[self.labels_ == i]
            else:
                X_i = X[:, self.labels_ == i]

            X_trans.append(self.strategy_(X_i, axis=self.data_dim_))

        if self.data_dim_ == 1:
            X_trans = np.stack(X_trans, axis=1)
        else:
            X_trans = np.array(X_trans)

        # Return based on vectorizes
        if not self.vectorize:
            return X_trans

        # Save original shape if vectorize called,
        # used for reverse transform
        self.original_transformed_shape_ = X_trans.shape
        
        return X_trans.flatten()

    def inverse_transform(self, X_trans):
        '''Reverse the original transformation.

        Parameters
        ----------
        X_trans : numpy.array
            Data with the same number of outputted
            features as data transformed
            by this object, e.g., the original
            transformed data or corresponding
            feature importances.

        Returns
        --------
        X : numpy.array
            The reverse transformed data
            passed back in its original space.

        '''
        
        # Fit pieces not depended on X if just
        # using object for inverse transform
        self._check_inverse_transform_fitted(X_trans)

        # Reverse the vectorize
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_transformed_shape_)

        X = np.zeros(self.X_shape_, dtype=X_trans.dtype, order='C')

        if self.data_dim_ == 1:
            X_trans = np.rollaxis(X_trans, -1)
            X = np.rollaxis(X, -1)

        for i, label in enumerate(self.non_bkg_unique_):
            X[self.labels_ == label] = X_trans[i]

        if self.data_dim_ == 1:
            X = np.rollaxis(X, -1)

        return X

    def _check_inverse_transform_fitted(self, X_trans):
        '''There is a case where we just want to use the Transformer Object
        to inverse transform, which means it has not been fitted yet.'''
        
        # If not fitted, fit based on reasonable values
        if not hasattr(self, "labels_"):
            self._base_fit()
            
            # Set just to whatever current shape is
            self.original_transformed_shape_ = X_trans.shape

            # Set to labels shape
            self.X_shape_ = self.labels_.shape

            # Data dim is 0 since the shapes match
            self.data_dim_ = 0


class SurfMaps(BaseEstimator, TransformerMixin):
    '''Extract signals from overlapping labels for surface data.

    This class functions similar to :class:`NiftiMapsMasker<nilearn.input_data.NiftiMapsMasker>`,
    except it is designed to work with surface projected data, instead of 3D/4D Nifti files.

    Parameters
    -----------
    maps : str or array-like, optional
        This parameter represents the maps in which
        to apply to each surface, where the shape of
        the passed maps should be (# of features, # of maps)
        or in other words, the size of the data array in the first
        dimension and the number of maps
        (i.e., the number of outputted ROIs from fit)
        as the second dimension.

        You may pass maps as either an array-like,
        or the str file location of a numpy or other
        valid surface file format array in which to load.
        Anything accepted by :func:`load<neurotools.loading.load>` is acceptable here.

    strategy : {'auto', 'ls', 'average'}, optional
        The strategy in which the maps are used to extract
        signal. If 'ls' is selected, which stands for
        least squares, the least-squares solution will
        be used for each region.

        Alternatively, if 'average' is passed, then
        the weighted average value for each map
        will be computed.

        By default 'auto' will be selected,
        which will use 'average' if the passed
        maps contain only positive weights, and
        'ls' in the case that there are
        any negative values in the passed maps.

        Otherwise, you can set a specific strategy.
        In deciding which method to use,
        consider an example. Let's say the
        fit data X, and maps are

        ::

            data = np.array([1, 1, 5, 5])
            maps = np.array([[0, 0],
                             [0, 0],
                             [1, -1],
                             [1, -1]])

        In this case, the 'ls' method would
        yield region signals [2.5, -2.5], whereas
        the weighted 'average' method, would yield
        [5, 5], notably ignoring the negative weights.
        This highlights an important limitation to the
        weighted averaged method, as it does not
        handle negative values well.

        On the other hand, consider changing the maps
        weights to

        ::

            data = np.array([1, 1, 5, 5])
            maps = np.array([[0, 1],
                             [0, 2],
                             [1, 0],
                             [1, 0]])

            ls_sol = [5. , 0.6]
            average_sol = [5, 1]

        In this case, we can see that the weighted
        average gives a maybe more intuitive summary
        of the regions. In general, it depends on
        what signal you are trying to summarize, and
        how you are trying to summarize it.

    mask : None, str or array-like, optional
        This parameter allows you to optional pass a mask of values in
        which to not calculate ROI values for.
        This can be passed as a str or
        array-like of values (just like maps),
        and should be comprised of
        a boolean array (or 1's and 0's),
        where a value of 1 means that value
        will be ignored (set to 0)
        should be kept, and a value of 0,
        for that value should be masked away.
        This array should have the same length as the passed `maps`.
        Specifically, where the shape of maps is (size, n_maps),
        the shape of mask should be (size).

        ::

            default = None

    vectorize : bool, optional
        If the returned array should be flattened to 1D. E.g., if this is
        the last step in a set of loader steps this should be True.
        Also note, if the surface data it is being applied to is 1D,
        then the output will be 1D regardless of this parameter.

        ::

            default = True

    See Also
    -----------
    SurfLabels : For extracting static / non-probabilistic parcellations.
    nilearn.input_data.NiftiMapsMasker : For volumetric nifti data.

    Examples
    ----------
    First let's define an example set of probabilistic maps, we
    will assume there are just 5 features in our data, and we will
    define 6 total maps.

    .. ipython:: python

        import numpy as np
        from neurotools.transform import SurfMaps

        # This should have shape number of features x number of maps!
        prob_maps = np.array([[3, 1, 1, 1, 1, 1],
                              [1, 3, 1, 1, 1, 1],
                              [1, 1, 3, 1, 1, 1],
                              [1, 1, 1, 3, 1, 1],
                              [1, 1, 1, 1, 3, 1]])
        prob_maps.shape

    Next we can define some input data to use with these maps.

    .. ipython:: python

        data1 = np.arange(5, dtype='float')
        data1
        data2 = np.ones(5, dtype='float')
        data2

    Now let's define the actual object and use it to transform the data.

    .. ipython:: python

        sm = SurfMaps(maps=prob_maps)
        sm.fit_transform(data1)
        sm.fit_transform(data2)

    Okay so what is going on when we transform this data? Basically we are
    just taking weighted averages for each one of the defined maps. We could
    also explicitly change the strategy from 'auto' to 'ls' which would
    take the least squares solution instead.

    .. ipython:: python

        sm = SurfMaps(maps=prob_maps, strategy='ls')
        data_trans = sm.fit_transform(data1)
        data_trans

    While a little less intuitive, the least squares solution allows
    us to reverse the feature transformation (although not always exactly)

    .. ipython:: python

        sm.inverse_transform(data_trans)

    This can be useful in the say the case of converting back downstream
    calculated feature importance to the original data space.
    '''

    def __init__(self, maps, strategy='auto', mask=None, vectorize=True):
        self.maps = maps
        self.strategy = strategy
        self.mask = mask
        self.vectorize = vectorize

    def fit(self, X, y=None):
        '''Fit this object according
        the passed subjects data, X.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None
        '''

        self._base_fit()

        # Save in dtype
        self.dtype_ = X.dtype

        # Warn if non-float
        if 'float' not in self.dtype_.name:
            warnings.warn('The original datatype is non-float, ' +
                          'this may lead to rounding errors! ' +
                          'Pass data as type float to ensure ' +
                          'the results of transform are not truncated.')

        # X can either be a 1D surface, or a 2D surface
        # (e.g. - for timeseries or stacked contrasts)
        if len(X.shape) > 2:
            raise RuntimeError('X can be at most 2D.')

        # Check to make sure dimension of data is correct
        if self.maps_.shape[0] not in X.shape:
            raise RuntimeError('Size of labels not found in X. '
                               'Make sure your data is in the same '
                               'space as the labels you are using!')

        return self

    def _base_fit(self):

        # Load mask if any
        self.mask_ = load(self.mask)

        # Load maps
        self.maps_ = load(self.maps)

        # Make the maps if passed mask set to 0 in those spots
        if self.mask_ is not None:

            # Raise error if wrong shapes
            if len(self.mask_) != self.maps_.shape[0]:
                raise RuntimeError('length of mask must have '
                                   'the same length / shape as '
                                   'the first dimension of passed '
                                   'maps!')
            self.maps_[self.mask_.astype(bool)] = 0

        # Make sure strategy exists
        if self.strategy not in ['auto', 'ls', 'average']:
            raise RuntimeError('strategy must be '
                               '"ls", "average" or "auto"!')

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform this object.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''

        return self.fit(X, y).transform(X)

    def _check_fitted(self):
        if not hasattr(self, "maps_"):
            raise ValueError('It seems that SurfMaps has not been fitted. '
                             'You must call fit() before calling transform()')

    def transform(self, X):
        '''Transform this the passed data.

        If X has the both the same dimension's, raise warning.

        Parameters
        ----------
        X : numpy.array
            numpy.array as either a 1D array,
            or a 2D array, where one dimension
            is matches the shape of the past
            labels and the other represents
            different time-points or modalities.

        y : numpy.array, optional
            This parameter is skipped.

            ::

                default = None

        Returns
        ---------
        X_trans : numpy.array
            The transformed data, either as 1D
            array is passed 1D data, or 2D if passed
            vectorize=False and originally 2D data.
        '''
        
        # If not fitted, do base fit
        if not hasattr(self, "maps_"):
            self._base_fit()

            # Default if not set
            self.dtype_ = 'float32'

        if len(X.shape) == 2 and (X.shape[0] == X.shape[1]):
            warnings.warn('X was passed with the same length',
                          ' in each dimension, ',
                          'Assuming that axis=0 is the data dimension',
                          ' w/ vertex values')

        # The data dimension is just the dimension with
        # the first dimension of maps
        self.data_dim_ = X.shape.index(self.maps_.shape[0])
        self.X_shape_ = X.shape

        # If data in second dimension, transpose
        if self.data_dim_ == 1:
            X = X.T

        # Set strategy if auto
        self.strategy_ = self.strategy
        if self.strategy_ == 'auto':
            self.strategy_ = 'ls'
            if np.all(self.maps_ >= 0):
                self.strategy_ = 'average'

        # Run the correct transform based on strategy
        if self.strategy_ == 'ls':
            X_trans = self._transform_ls(X)
        elif self.strategy_ == 'average':
            X_trans = self._transform_average(X)
        else:
            X_trans = None

        # Convert back to original dtype
        X_trans = X_trans.astype(self.dtype_)

        # Always return as shape of extra data if any by
        # number of maps, with number of maps as last dimension

        # Return based on vectorize
        if not self.vectorize:
            return X_trans

        # Save original transformed output shape if vectorize
        self.original_transformed_shape_ = X_trans.shape
        return X_trans.flatten()

    def _transform_ls(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = lstsq(self.maps_, X)[0]
        return X_trans.T

    def _transform_average(self, X):
        '''X should be data points, or data points x stacked.'''

        X_trans = []

        # For each map - take weighted average
        for m in range(self.maps_.shape[1]):

            try:
                X_trans.append(np.average(X, axis=0, weights=self.maps_[:, m]))
            except ZeroDivisionError:
                pass

        return np.array(X_trans).T

    def inverse_transform(self, X_trans):
        '''Reverse the original transformation.

        Parameters
        ----------
        X_trans : numpy.array
            Data with the same number of outputted
            features as data transformed
            by this object, e.g., the original
            transformed data or corresponding
            feature importances.

        Returns
        --------
        X : numpy.array
            The reverse transformed data
            passed back in its original space.

        '''

        # Reverse the vectorize, if needed
        if self.vectorize:
            X_trans = X_trans.reshape(self.original_transformed_shape_)

        if self.strategy_ == 'ls':
            return np.dot(X_trans, self.maps_.T)

        elif self.strategy_ == 'average':
            raise RuntimeError('Cannot calculate reverse of average.')


def project_map_fis(fis, atlas, label_names=None):

    # Get values as array
    atlas_vals = atlas.get_fdata()
    
    # Grab example slice to init
    ex = atlas.slicer[:, :, :, 0]
    fill = np.zeros(ex.shape)
    
    # Fill each
    for i in range(len(fis)):
        
        fi = fis[i]
    
        # If passed series, use label names to index
        if isinstance(fis, pd.Series):
            map_i = np.where(label_names == fis.index[i])[0][0]
        else:
            map_i = i
        
        # Get reference ROI
        roi = atlas_vals[:, :, :, map_i]

        # Distribute weight based on roi map
        fill += (roi * fi) / roi.sum()
    
    # Return in nifti image
    return nib.Nifti1Image(fill, affine=ex.affine)


import pandas as pd
import re

def name_replace(in_obj, to_replace, replace_with):
    '''Helper function to perform simple text replacement but
    over arbitrary objects.

    Parameters
    -----------
    in_obj : list, set or :class:`pandas.DataFrame`
        The object in which to replace all sub objects of,
        or in the case of DataFrame input the column names of.
        
        Note: This replacement occurs in some cases in-place rather
        than on a copy of the object (except in the case of set input).

    to_replace : str or list of str
        The str in which to replace with other argument replace_with.
        In the case that a list of str is passed, then
        any of the elements of the passed list will be if found replaced with
        the passed argument for replace_with.

    replace_with : str
        The str in which all instances of to_replace should be replaced with.

    Returns
    --------
    out_obj : list, set or :class:`pandas.DataFrame`
        The in_obj but with text replacement performed / column name replacement.


    Examples
    ----------

    .. ipython:: python

        import pandas as pd
        from neurotools.misc.text import name_replace
        
        # Base case
        name_replace(['cow123', 'goat123', 'somethingelse'], '123', '456')
        
        # DataFrame case
        df = pd.DataFrame(columns=['cow123', 'goat123', 'somethingelse'])
        name_replace(df, '123', '456')
        
        # to_replace list case
        name_replace(in_obj=['cow123', 'goat123', 'somethingelse'],
                     to_replace=['123', 'something'],
                     replace_with='456')

    '''
    
    # List case
    if isinstance(in_obj, list):
        for i in range(len(in_obj)):
            if isinstance(to_replace, list):
                for tr in to_replace:
                    in_obj[i] = in_obj[i].replace(tr, replace_with)
            else:
                in_obj[i] = in_obj[i].replace(to_replace, replace_with)

        return in_obj

    # Set case
    if isinstance(in_obj, set):

        new_obj = set()
        for item in in_obj:
            new_item = item

            if isinstance(to_replace, list):
                for tr in to_replace:
                    new_item = new_item.replace(tr, replace_with)
            else:
                new_item = new_item.replace(to_replace, replace_with)

            new_obj.add(new_item)

        return new_obj
    
    # DataFrame case
    if isinstance(in_obj, pd.DataFrame):

        cols = list(in_obj)
        col_name_replace = {}

        for col in cols:
            if isinstance(to_replace, list):
                for tr in to_replace:
                    if tr in col:
                        col_name_replace[col] = col.replace(tr, replace_with)
            else:
                if to_replace in col:
                    col_name_replace[col] = col.replace(to_replace, replace_with)

        return in_obj.rename(col_name_replace, axis=1)

    raise RuntimeError(f'type of {in_obj} not supported.')

def readable_size_to_bytes(size):
    '''Based on https://stackoverflow.com/questions/42865724/parse-human-readable-filesizes-into-bytes'''
    
    # Assume already bytes if fails convert to float
    try:
        return float(size)
    except ValueError:
        pass

    # Process as str, with 2 versions
    units = {"B": 1,
             "KB": 2**10, "K": 2**10,
             "MB": 2**20, "M": 2**20,
             "GB": 2**30, "G": 2**30,
             "TB": 2**40, "T": 2**40}
    
    # Make sure all uppercase
    size = size.upper()

    # Match replace with or w/o B
    if not re.match(r' ', size):
        size = re.sub(r'([KMGT]B?)', r' \1', size)

    # Split number and unit
    number, unit = [string.strip() for string in size.split()]

    # Return number * unit as int
    return int(float(number)*units[unit])

def substrs(x):
    return {x[i:i+j] for i in range(len(x)) for j in range(len(x) - i + 1)}

def find_substr(data):

    s = substrs(data[0])

    for val in data[1:]:
        s.intersection_update(substrs(val))

    try:
        mx = max(s, key=len)

    except ValueError:
        mx = ''

    return mx

def get_top_substrs(keys):

    found = []
    top = find_substr(keys)

    while len(top) > 1:
        found.append(top)

        keys = [k.replace(top, '') for k in keys]
        top = find_substr(keys)

    return found

def get_unique_str_markers(strs):

    # Get all top common sub strings
    top_substrs = get_top_substrs(strs)

    # Remove all common strs from each individual str
    unique_strs = []
    for s in strs:
        piece = s
        for substr in top_substrs:
            piece = piece.replace(substr, '')

        # Add final
        unique_strs.append(piece)

    return unique_strs

def clean_key(key):
    '''Applies lowercase, and some common str replacements, "cleaning" the key.'''
    
    # TODO use regex ? 
    key = key.lower()
    key = key.replace('.', ' ')
    key = key.replace('-', ' ')
    key = key.replace('_', ' ')
    key = key.replace('  ', ' ')
    key = key.replace('   ', ' ')
    key = key.replace('\t', ' ')
    key = key.replace('\n', ' ')

    key = key.replace('&', ' and ')
    
    return key


from sklearn.base import BaseEstimator
import numpy as np


class OrthRegression(BaseEstimator):
    
    def __init__(self):
        pass
    
    def fit(self, X, y):
        
        # Work with 1D
        x = np.squeeze(X)
        
        # Concat
        Z  = np.c_[x, y]
        
        # Compute SVD
        meanZ = np.tile(Z.mean(axis=0), (Z.shape[0], 1))
        V = np.linalg.svd(Z - meanZ)[2]

        # Slope
        self.a = -V[1][0] / V[1][1]

        # Coef
        self.b = np.mean(np.matmul(Z, V[1]) / V[1, 1])

        # Save normal for use w/ get orth dist
        self.normal = V[1]

        return self
        
    def predict(self, X):
        
        # Work with 1D x
        x = np.squeeze(X)
        
        # Get resid
        resid = (x * self.a) + self.b
        
        # Return as len(x) x 1
        return np.expand_dims(resid, -1)
        
    def get_orth_dist(self, X, y):

        # Get z
        x = np.squeeze(X)
        Z  = np.c_[x, y]
        meanZ = np.tile(Z.mean(axis=0), (Z.shape[0], 1))

        # Compute orth distances
        return np.abs(np.matmul((Z - meanZ), self.normal))


from ..space import process_space
import numpy as np

def test_basic_arr_inputs():

    # Fsaverage 3, one hemi, needs medial test case
    proc_data, space = process_space(np.zeros(578), hemi='lh')
    assert proc_data['lh'].shape == (642, )
    assert 'rh' not in proc_data
    assert 'vol' not in proc_data
    assert 'sub' not in proc_data
    assert space == 'fsaverage3'



from ..measures import ThresholdNetworkMeasures
import numpy as np

def test_threshold_network_measures_abs():

    nm =\
        ThresholdNetworkMeasures(threshold=0.2,
                                 threshold_type='abs',
                                 threshold_method='value',
                                 to_compute='avg_degree')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    # Fit
    nm.fit_transform(X)
    assert nm.feat_names_[0] == 'avg_degree'
    assert len(nm.feat_names_) == 1

    # Threshold should stay fixed
    nm.threshold_check(X)
    assert nm.threshold == .2

    thresh_X = nm._apply_threshold(X)
    assert len(np.unique(thresh_X)) == 2

    # Should just drop .1 edge
    assert np.sum(thresh_X) == 8

    import networkx as nx
    G = nx.from_numpy_array(thresh_X)

    degrees = [n[1] for n in G.degree()]
    X_trans = nm.transform(X)
    assert len(X_trans) == 1
    assert np.mean(degrees) == X_trans[0]


def test_threshold_network_measures_neg():

    nm =\
        ThresholdNetworkMeasures(threshold=-.2,
                                 threshold_type='neg',
                                 threshold_method='value')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
    assert np.sum(thresh_X) == 4
    assert thresh_X[0][-1] == 1
    assert thresh_X[0][1] == 0


def test_threshold_network_measures_pos():


    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='pos',
                                 threshold_method='value')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
    assert np.sum(thresh_X) == 4
    assert thresh_X[0][-1] == 0
    assert thresh_X[0][1] == 1


def test_threshold_network_measures_density():

    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='abs',
                                 threshold_method='density')

    X = np.array([[0, .2, .3, -.4],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.4, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
   
    # 20% of 16 possible edges
    # rounded up is 4
    assert np.sum(thresh_X) == 4


def test_threshold_network_measures_density_pos():

    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='pos',
                                 threshold_method='density')

    X = np.array([[0, .2, .3, -.4],
                  [.2, 0, -.5, 0],
                  [0, -.5, 0, 0],
                  [-.4, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
   
    # 20% of 16 possible edges
    # rounded up is 4
    assert np.sum(thresh_X) == 3


def test_threshold_network_measures_density_neg():

    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='neg',
                                 threshold_method='density')

    X = np.array([[0, .2, .3, -.4],
                  [.2, 0, -.5, 0],
                  [.3, 0, 0, 0],
                  [-.4, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
   
    # 20% of 16 possible edges
    # rounded up is 4
    assert np.sum(thresh_X) == 3



from sklearn.linear_model import LinearRegression
import numpy as np

def get_resid(covars, data):
    '''Compute the simple linear residualized version of a set of data according
    to a passed set of covariates. This fits a linear regression from covariates to data
    and then returns the difference between the predicted values and the true values for
    the data.

    In this case of missing data, within the data portion of the input,
    this method will operate on only the the subset of subject's with
    non-missing data per feature, any NaN data will be propegated to
    the returned residualized data.

    Note that the intercept information from the linear model
    is re-added to the residuals. 

    Parameters
    ------------
    covars : numpy array
        The covariates in which to use to residualize
        the passed data, seperate per feature. This
        input must have shape number of subjects x number of
        covariates.

    data : numpy array
        The data in which to residualize. This
        input muct have shape number of subjects x number of
        data features

    Returns
    ---------
    resid_data : numpy array
        A residualized version, with the same shape,
        of the passed data is returned. Any NaN values
        in the original input data are preserved.
    '''

    if np.isnan(data).any():
        return _get_resid_with_nans(covars, data)
    return _get_resid_without_nans(covars, data)

def _get_resid_with_nans(covars, data):

    # Go from pandas df to numpy array
    covars = np.array(covars)

    # Make sure data is numpy array
    data = np.array(data)
    
    # Init empty resid array of NaN's
    resid = np.empty(shape=data.shape)
    resid[:] = np.nan
    
    # For each feature seperately
    for i in range(data.shape[1]):
        
        # Operate on non-nan subjects for this feature
        mask = ~np.isnan(data[:, i])
        
        # If not at least 2 subjects valid,
        # skip and propegate NaN's for this
        # voxel.
        if np.sum(mask) > 1:
        
            # Fit model
            model = LinearRegression().fit(covars[mask], data[mask, i])
            
            # Compute difference of real value - predicted, then add intercept
            resid_i = data[mask, i] - model.predict(covars[mask]) + model.intercept_
            
            # Fill in NaN mask
            resid[mask, i] = resid_i
    
    return resid

def _get_resid_without_nans(covars, data):

    # Go from pandas df to numpy array
    covars = np.array(covars)

    # Make sure data is numpy array
    data = np.array(data)

    # Fit a linear regression with covars as predictors, each voxel as target variable
    model = LinearRegression().fit(covars, data)

    # The difference is the real value of the voxel, minus the predicted value, + intercepts
    resid = data - model.predict(covars) + model.intercept_

    return resid

def get_cohens(data):
    '''Get cohen's d value with or without any
    missing values present in data'''

    if np.isnan(data).any():
        return _get_cohens_with_nan(data)
    return _get_cohens_without_nans(data)

def _get_cohens_with_nan(data):

    mean = np.nanmean(data, axis=0)
    std = np.nanstd(data, axis=0)
    cohen = mean / std

    return cohen

def _get_cohens_without_nans(data):

    mean = np.mean(data, axis=0)
    std = np.std(data, axis=0)
    cohen = mean / std

    return cohen

def fast_corr(O, P):
    '''Compute a very fast correlation coef.'''
    
    n = P.size
    DO = O - (np.sum(O, 0) / np.double(n))
    DP = P - (np.sum(P) / np.double(n))
    return np.dot(DP, DO) / np.sqrt(np.sum(DO ** 2, 0) * np.sum(DP ** 2))


from nibabel.gifti import gifti
import numpy as np
from nibabel.freesurfer.io import read_geometry

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.surface import load_surf_mesh

def geo_to_gifti(geo_coords, geo_faces):

    coords = gifti.GiftiDataArray(data=np.asarray(geo_coords, 'float32'),
                                  intent=gifti.intent_codes.field1['pointset'],
                                  datatype=16,
                                  endian=2,
                                  coordsys=gifti.GiftiCoordSystem(dataspace=3, xformspace=3))

    faces = gifti.GiftiDataArray(data=np.asarray(geo_faces, np.int32),
                                 intent=gifti.intent_codes.field1['triangle'],
                                 datatype=8,
                                 endian=2,
                                 coordsys=gifti.GiftiCoordSystem(dataspace=3, xformspace=3))

    img = gifti.GiftiImage()
    img.add_gifti_data_array(coords)
    img.add_gifti_data_array(faces)
    
    return img

def data_to_gifti(s_data):
    
    data = gifti.GiftiDataArray(data=np.asarray(s_data, 'float32'),
                                intent=11,
                                datatype=16,
                                endian=2)
    
    img = gifti.GiftiImage()
    img.add_gifti_data_array(data)
    
    return img

def load_geo_as_network(geo):
    '''Quick utility for loading a geometry file / just
    set of triangles as a networkx network.'''
    
    if isinstance(geo, 'str'):
        
        try:
            _, geo = load_surf_mesh(geo)
        except ValueError:
            _, geo = read_geometry(geo)
    
    # If passed as tuple of mesh
    elif len(geo) == 2:
        geo = geo[1]
        
    G = nx.Graph()
    
    for tri in geo:
        G.add_edge(tri[0], tri[1])
        G.add_edge(tri[0], tri[2])
        G.add_edge(tri[1], tri[2])
    
    return G
    



import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from matplotlib.colorbar import make_axes
from matplotlib.cm import ScalarMappable, get_cmap
from matplotlib.colors import Normalize, LinearSegmentedColormap
from distutils.version import LooseVersion
import matplotlib

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.surface import load_surf_data, load_surf_mesh

def _crop_colorbar(cbar, cbar_vmin, cbar_vmax):
    """Crop a colorbar to show from cbar_vmin to cbar_vmax.
    Used when symmetric_cbar=False is used.
    """
    if (cbar_vmin is None) and (cbar_vmax is None):
        return
    cbar_tick_locs = cbar.locator.locs
    if cbar_vmax is None:
        cbar_vmax = cbar_tick_locs.max()
    if cbar_vmin is None:
        cbar_vmin = cbar_tick_locs.min()
    new_tick_locs = np.linspace(cbar_vmin, cbar_vmax,
                                len(cbar_tick_locs))

    # matplotlib >= 3.2.0 no longer normalizes axes between 0 and 1
    # See https://matplotlib.org/3.2.1/api/prev_api_changes/api_changes_3.2.0.html
    # _outline was removed in
    # https://github.com/matplotlib/matplotlib/commit/03a542e875eba091a027046d5ec652daa8be6863
    # so we use the code from there
    if LooseVersion(matplotlib.__version__) >= LooseVersion("3.2.0"):
        cbar.ax.set_ylim(cbar_vmin, cbar_vmax)
        X = cbar._mesh()[0]
        X = np.array([X[0], X[-1]])
        Y = np.array([[cbar_vmin, cbar_vmin], [cbar_vmax, cbar_vmax]])
        N = X.shape[0]
        ii = [0, 1, N - 2, N - 1, 2 * N - 1, 2 * N - 2, N + 1, N, 0]
        x = X.T.reshape(-1)[ii]
        y = Y.T.reshape(-1)[ii]
        xy = (np.column_stack([y, x])
              if cbar.orientation == 'horizontal' else
              np.column_stack([x, y]))
        cbar.outline.set_xy(xy)
    else:
        cbar.ax.set_ylim(cbar.norm(cbar_vmin), cbar.norm(cbar_vmax))
        outline = cbar.outline.get_xy()
        outline[:2, 1] += cbar.norm(cbar_vmin)
        outline[2:6, 1] -= (1. - cbar.norm(cbar_vmax))
        outline[6:, 1] += cbar.norm(cbar_vmin)
        cbar.outline.set_xy(outline)

    cbar.set_ticks(new_tick_locs)
    cbar.update_ticks()

VALID_VIEWS = "anterior", "posterior", "medial", "lateral", "dorsal", "ventral"
VALID_HEMISPHERES = "left", "right"

MATPLOTLIB_VIEWS = {"right": {"lateral": (0, 0),
                              "medial": (0, 180),
                              "dorsal": (90, 0),
                              "ventral": (270, 0),
                              "anterior": (0, 90),
                              "posterior": (0, 270)
                              },
                    "left": {"medial": (0, 0),
                             "lateral": (0, 180),
                             "dorsal": (90, 0),
                             "ventral": (270, 0),
                             "anterior": (0, 90),
                             "posterior": (0, 270)
                             }
                    }

def _set_view_plot_surf_matplotlib(hemi, view):
    """Helper function for plot_surf with matplotlib engine.
    This function checks the selected hemisphere and view, and
    returns elev and azim.
    """
    if hemi not in VALID_HEMISPHERES:
        raise ValueError(f"hemi must be one of {VALID_HEMISPHERES}")
    if view not in MATPLOTLIB_VIEWS[hemi]:
        raise ValueError(f"view must be one of {VALID_VIEWS}")
    return MATPLOTLIB_VIEWS[hemi][view]




def plot_single_surf(surf_mesh, surf_map=None, bg_map=None,
                     hemi='left', view='lateral', cmap=None, colorbar=False,
                     avg_method='mean', threshold=None, alpha='auto',
                     bg_on_data=False, darkness=1, vmin=None, vmax=None,
                     cbar_vmin=None, cbar_vmax=None,
                     title=None, output_file=None, axes=None, figure=None,
                     dist=8, cbar_tick_format='%.2g', **kwargs):

    _default_figsize = [4, 4]

    # load mesh and derive axes limits
    mesh = load_surf_mesh(surf_mesh)
    coords, faces = mesh[0], mesh[1]
    limits = [coords.min(), coords.max()]

    # Get elev and azim from view
    elev, azim = _set_view_plot_surf_matplotlib(hemi, view)

    # set alpha if in auto mode
    if alpha == 'auto':
        if bg_map is None:
            alpha = .5
        else:
            alpha = 1

    # if no cmap is given, set to matplotlib default
    if cmap is None:
        cmap = plt.cm.get_cmap(plt.rcParamsDefault['image.cmap'])
    else:
        # if cmap is given as string, translate to matplotlib cmap
        if isinstance(cmap, str):
            cmap = plt.cm.get_cmap(cmap)

    figsize = _default_figsize
    
    # Leave space for colorbar
    if colorbar:
        figsize[0] += .7
    
    # Initiate figure and 3d axes
    if axes is None:
        if figure is None:
            figure = plt.figure(figsize=figsize)
        axes = figure.add_axes((0, 0, 1, 1), projection="3d")
    else:
        if figure is None:
            figure = axes.get_figure()

    axes.set_xlim(*limits)
    axes.set_ylim(*limits)
    axes.view_init(elev=elev, azim=azim)
    axes.set_axis_off()

    # plot mesh without data
    p3dcollec = axes.plot_trisurf(coords[:, 0], coords[:, 1], coords[:, 2],
                                  triangles=faces, linewidth=0.,
                                  antialiased=False,
                                  color='white')

    # reduce viewing distance to remove space around mesh
    axes.dist = dist

    # set_facecolors function of Poly3DCollection is used as passing the
    # facecolors argument to plot_trisurf does not seem to work
    face_colors = np.ones((faces.shape[0], 4))

    if bg_map is None:
        bg_data = np.ones(coords.shape[0]) * 0.5

    else:
        bg_data = load_surf_data(bg_map)
        if bg_data.shape[0] != coords.shape[0]:
            raise ValueError('The bg_map does not have the same number '
                             'of vertices as the mesh.')

    bg_faces = np.mean(bg_data[faces], axis=1)
    if bg_faces.min() != bg_faces.max():
        bg_faces = bg_faces - bg_faces.min()
        bg_faces = bg_faces / bg_faces.max()
    
    # Control background darkness
    bg_faces *= darkness
    face_colors = plt.cm.gray_r(bg_faces)

    # modify alpha values of background
    face_colors[:, 3] = alpha * face_colors[:, 3]
    # should it be possible to modify alpha of surf data as well?

    if surf_map is not None:
        surf_map_data = load_surf_data(surf_map)
        
        if len(surf_map_data.shape) != 1:
            raise ValueError('surf_map can only have one dimension but has'
                             '%i dimensions' % len(surf_map_data.shape))
        if surf_map_data.shape[0] != coords.shape[0]:
            raise ValueError('The surf_map does not have the same number '
                             'of vertices as the mesh.')

        # create face values from vertex values by selected avg methods
        if avg_method == 'mean':
            surf_map_faces = np.mean(surf_map_data[faces], axis=1)
        elif avg_method == 'median':
            surf_map_faces = np.median(surf_map_data[faces], axis=1)
        elif avg_method == 'min':
            surf_map_faces = np.min(surf_map_data[faces], axis=1)
        elif avg_method == 'max':
            surf_map_faces = np.max(surf_map_data[faces], axis=1)
        elif callable(avg_method):
            surf_map_faces =\
                np.apply_along_axis(avg_method, 1, surf_map_data[faces])

            ## check that surf_map_faces has the same length as face_colors
            if surf_map_faces.shape != (face_colors.shape[0],):
                raise ValueError(
                    'Array computed with the custom function '
                    'from avg_method does not have the correct shape: '
                    '{} != {}'.format(
                        surf_map_faces.shape[0],
                        face_colors.shape[0]
                    )
                )

            ## check that dtype is either int or float
            if not (
                "int" in str(surf_map_faces.dtype) or
                "float" in str(surf_map_faces.dtype)
            ):
                raise ValueError(
                    'Array computed with the custom function '
                    'from avg_method should be an array of numbers '
                    '(int or float)'
                )
        else:
            raise ValueError(
                "avg_method should be either "
                "['mean', 'median', 'max', 'min'] "
                "or a custom function"
            )

        # if no vmin/vmax are passed figure them out from data
        if vmin is None:
            vmin = np.nanmin(surf_map_faces)
        if vmax is None:
            vmax = np.nanmax(surf_map_faces)

        # Threshold if indicated
        if threshold is None:

            # If no thresholding and nans, filter them out
            kept_indices = np.where(
                            np.logical_not(
                                np.isnan(surf_map_faces)))[0]
        else:
            kept_indices = np.where(np.abs(surf_map_faces) >= np.abs(threshold))[0]

        surf_map_faces = surf_map_faces - vmin
        surf_map_faces = surf_map_faces / (vmax - vmin)

        # multiply data with background if indicated
        if isinstance(bg_on_data, bool):
            if bg_on_data:
                face_colors[kept_indices] = cmap(surf_map_faces[kept_indices])\
                    * face_colors[kept_indices]
            else:
                face_colors[kept_indices] = cmap(surf_map_faces[kept_indices])
        else:
            face_colors[kept_indices] = cmap(surf_map_faces[kept_indices])\
                    * plt.cm.gray_r(bg_faces * bg_on_data)[kept_indices]

        if colorbar:
            our_cmap = get_cmap(cmap)
            norm = Normalize(vmin=vmin, vmax=vmax)

            # Default number of ticks is 5...
            nb_ticks = 5
            # ...unless we are dealing with integers with a small range
            # in this case, we reduce the number of ticks
            if cbar_tick_format == "%i" and vmax - vmin < nb_ticks:
                ticks = np.arange(vmin, vmax + 1)
            else:
                ticks = np.linspace(vmin, vmax, nb_ticks)

            bounds = np.linspace(vmin, vmax, our_cmap.N)

            if threshold is not None:
                cmaplist = [our_cmap(i) for i in range(our_cmap.N)]
                # set colors to grey for absolute values < threshold
                istart = int(norm(-threshold, clip=True) *
                             (our_cmap.N - 1))
                istop = int(norm(threshold, clip=True) *
                            (our_cmap.N - 1))
                for i in range(istart, istop):
                    cmaplist[i] = (0.5, 0.5, 0.5, 1.)
                our_cmap = LinearSegmentedColormap.from_list(
                    'Custom cmap', cmaplist, our_cmap.N)

            # we need to create a proxy mappable
            proxy_mappable = ScalarMappable(cmap=our_cmap, norm=norm)
            proxy_mappable.set_array(surf_map_faces)
            cax, kw = make_axes(axes, location='right', fraction=.15,
                                shrink=.5, pad=.0,  aspect=10.)
            cbar = figure.colorbar(
                proxy_mappable, cax=cax, ticks=ticks,
                boundaries=bounds, spacing='proportional',
                format=cbar_tick_format, orientation='vertical')
            _crop_colorbar(cbar, cbar_vmin, cbar_vmax)

        p3dcollec.set_facecolors(face_colors)
    
    # Try set background 0
    axes.patch.set_alpha(0)

    if title is not None:
        axes.set_title(title, position=(.5, .95))

    # save figure if output file is given
    if output_file is not None:
        figure.savefig(output_file)
        plt.close(figure)
    else:
        return figure, axes, surf_map_faces


def add_collage_colorbar(figure, ax, smfs, vmin, vmax, 
                         cbar_vmin=None, cbar_vmax=None,
                         multicollage=False,
                         colorbar_params={},
                         **kwargs):

    if 'fraction' in colorbar_params:
        fraction=colorbar_params['fraction']
    else:
        fraction = .5

    if 'shrink' in colorbar_params:
        shrink=colorbar_params['shrink']
    else:
        shrink = .5

    if 'aspect' in colorbar_params:
        aspect=colorbar_params['aspect']
    else:
        aspect = 20

    if 'pad' in colorbar_params:
        pad=colorbar_params['pad']
    else:
        pad = .5

    if 'anchor' in colorbar_params:
        anchor = colorbar_params['anchor']
    else:
        anchor = (0.0, 0.5)

    if 'format' in colorbar_params:
        c_format = colorbar_params['format']
    else:
        c_format = '%.2g'

    if 'cmap' not in kwargs:
        cmap = None
    else:
        cmap = kwargs.pop('cmap')

    if cmap is None:
        cmap = get_cmap(plt.rcParamsDefault['image.cmap'])
    else:
        if isinstance(cmap, str):
            cmap = get_cmap(cmap)

    if 'threshold' in kwargs:
        threshold = kwargs['threshold']
    else:
        threshold = None

    # Color bar
    our_cmap = get_cmap(cmap)

    norm = Normalize(vmin=vmin, vmax=vmax)

    nb_ticks = 5
    ticks = np.linspace(vmin, vmax, nb_ticks)
    bounds = np.linspace(vmin, vmax, our_cmap.N)

    if threshold is not None:
        cmaplist = [our_cmap(i) for i in range(our_cmap.N)]

        # set colors to grey for absolute values < threshold
        istart = int(norm(-threshold, clip=True) *
                     (our_cmap.N - 1))
        istop = int(norm(threshold, clip=True) *
                    (our_cmap.N - 1))
        for i in range(istart, istop):
            cmaplist[i] = (0.5, 0.5, 0.5, 1.)
        our_cmap = LinearSegmentedColormap.from_list(
            'Custom cmap', cmaplist, our_cmap.N)

    # we need to create a proxy mappable
    proxy_mappable = ScalarMappable(cmap=our_cmap, norm=norm)
    proxy_mappable.set_array(smfs)

    if multicollage:

        cbar = plt.colorbar(
            proxy_mappable, ax=ax, ticks=ticks, spacing='proportional',
            format=c_format, orientation='vertical',
            anchor=anchor, boundaries=bounds,
            fraction=fraction,
            shrink=shrink,
            aspect=aspect,
            pad=pad)

    else:

        left = (ax[0][0].get_position().x0 + ax[0][0].get_position().x1) / 2
        right = (ax[0][1].get_position().x0 + ax[0][1].get_position().x1) / 2
        bot = ax[1][0].get_position().y1
        width = right-left

        # [left, bottom, width, height]
        cbaxes = figure.add_axes([left, bot - (.05 / 3), width, .05])

        cbar = plt.colorbar(
            proxy_mappable, cax=cbaxes, ticks=ticks, spacing='proportional',
            format=c_format, orientation='horizontal', shrink=1, anchor='C')

    _crop_colorbar(cbar, cbar_vmin, cbar_vmax)

    return figure, ax
